#!/usr/bin/python
"""ノード定義

    * UserとGroupのモデル定義を書きます。
    * 関係テーブルのモデル実装は別モジュールにしようかと思ってる
    * sqlalchemyのベースクラスを拡張したNodeクラスに共通のプロパティを載せて、そいつらをUserとGroupに継承させてます。

Todo:
    * sqlalchemy用にUser型とGroup型を作って、↓のクラスをそのまま使ってDB呼び出しできるようにしたい
"""

from .core_db import db
from .nodes import (
    User,
    Group,
    Node,
    UserGroup,
    UserInvite,
    AuthorizationError,
    Role,
    InnerRole,
    PushSubscription
)
from .permissions import permission, inner_permission


Permission = permission(db, [
    {
        "name": "権限管理",
        "screen_name": "admin"
    },
    {
        "name": "グループ作成",
        "screen_name": "create_group",
    },
    {
        "name": "グループ更新",
        "screen_name": "update_group",
    },
    {
        "name": "グループ削除",
        "screen_name": "delete_group",
    },
    {
        "name": "ユーザー作成",
        "screen_name": "create_user",
    },
    {
        "name": "ユーザー更新",
        "screen_name": "update_user",
    },
    {
        "name": "ユーザー削除",
        "screen_name": "delete_user",
    }
])

InnerPermission = inner_permission(db, [
    {
        "name": "グループ管理",
        "screen_name": "admin"
    },
    {
        "name": "ユーザー追加",
        "screen_name": "add_user",
    },
    {
        "name": "ユーザー削除",
        "screen_name": "remove_user"
    },
    {
        "name": "グループ追加",
        "screen_name": "add_group"
    },
    {
        "name": "グループ削除",
        "screen_name": "remove_group"
    }
])


def is_admin(node):
    return Permission.is_accepted('admin', node)


db.create_all()


__all__ = [
    User,
    Group,
    Node,
    UserGroup,
    UserInvite,
    AuthorizationError,
    Role,
    InnerRole,
    Permission,
    InnerPermission,
    PushSubscription,
    permission,
    inner_permission,
]



import importlib
import os
import sys
from pathlib import Path
import json

from watchdog.events import FileSystemEvent, PatternMatchingEventHandler
from watchdog.observers import Observer

from mitama._extra import _Singleton
from mitama.app.http import Request, Response
from mitama.app.app import _session_middleware

from .method import group
from .router import Router


class AppRegistry(_Singleton):
    """稼働しているアプリのレジストリ

    サーバー内で稼働しているアプリのパスやパッケージ名が登録されているレジストリです。
    mitama.jsonを読み込んでアプリを起動するクラスでもあります。
    dictっぽくアプリの取得や配信の停止などが可能です。
    """

    _map = dict()
    _server = None
    _router = None

    def __init__(self):
        super().__init__()

    def __iter__(self):
        for app in self._map.values():
            yield app

    def __setitem__(self, app_name, app):
        app.project = self.project
        self._map[app_name] = app

        def sorter(x):
            x_ = x[1].path
            if x_[-1] != "/":
                x_ += "/"
            return -1 * len(x_.split('/'))

        self._map = dict(
            sorted(
                self._map.items(),
                key=sorter
            )
        )

    def __getitem__(self, app_name):
        return self._map[app_name]

    def __delitem__(self, app_name):
        del self._map[app_name]

    def reset(self):
        """アプリの一覧をリセットします"""
        self._map = dict()

    def load_package(self, package, screen_name=None, path="/"):
        project_dir = self.project.project_dir
        if str(project_dir) not in sys.path:
            sys.path.append(str(project_dir))
        if app_name not in sys.modules:
            init = importlib.__import__(app_name, fromlist=["AppBuilder"])
        else:
            init = importlib.reload(app_name)
        builder = init.AppBuilder()
        if screen_name is None:
            screen_name = package
        builder.set_package(package)
        builder.set_screen_name(screen_name)
        builder.set_project_root_dir(project_dir)
        builder.set_project_dir(project_dir / screen_name)
        builder.set_path(path)
        app = builder.build()
        return app

    def router(self):
        """アプリの情報に基づいてルーティングエンジンを生成します"""
        if self._router == None:
            router = Router(
                middlewares=[_session_middleware()]
            )
            for app in self:
                router.add_route(group(app.path, app))
            self._router = router
        return self._router



def _session_middleware():
    import base64

    from Crypto.Random import get_random_bytes

    from mitama.app import Middleware
    from mitama.app.http.session import EncryptedCookieStorage

    if "MITAMA_SESSION_KEY" in os.environ:
        session_key = os.environ["MITAMA_SESSION_KEY"]
    elif os.path.exists(".tmp/MITAMA_SESSION_KEY"):
        with open(".tmp/MITAMA_SESSION_KEY", "r") as f:
            session_key = f.read()
    else:
        key = get_random_bytes(16)
        session_key = base64.urlsafe_b64encode(key).decode("utf-8")
        if not os.path.exists(".tmp"):
            os.mkdir(".tmp")
        with open(".tmp/MITAMA_SESSION_KEY", "w") as f:
            f.write(session_key)

    class SessionMiddleware(Middleware):
        fernet_key = session_key

        def __init__(self):
            secret_key = base64.urlsafe_b64decode(self.fernet_key.encode("utf-8"))
            cookie_storage = EncryptedCookieStorage(secret_key)
            self.storage = cookie_storage

        def process(self, request, handler):
            request["mitama_session_storage"] = self.storage
            raise_response = False
            response = handler(request)
            if not isinstance(response, Response):
                return response
            session = request.get("mitama_session")
            if session is not None:
                if session._changed:
                    self.storage.save_session(request, response, session)
            if raise_response:
                raise response
            return response

    return SessionMiddleware




from mitama.db import DatabaseManager, ForeignKey, relationship, Table
from mitama.db.types import Column
from mitama.db.types import String
from sqlalchemy import event


def permission(db_, permissions):
    from .nodes import User, Group, UserGroup, Role, InnerRole
    role_permission = Table(
        db_.Model.prefix + "_role_permission",
        db_.metadata,
        Column(
            "role_id",
            String(64),
            ForeignKey("mitama_role._id", ondelete="CASCADE"),
            primary_key=True
        ),
        Column(
            "permission_id",
            String(64),
            ForeignKey(
                db_.Model.prefix + "_permission._id",
                ondelete="CASCADE"
            ),
            primary_key=True
        ),
        extend_existing=True
    )

    class Permission(db_.Model):
        name = Column(String(64))
        screen_name = Column(String(64), unique=True)
        roles = relationship(
            "Role",
            secondary=role_permission,
            cascade="all, delete"
        )

        @classmethod
        def accept(cls, screen_name, role):
            """特定のRoleに許可します """
            permission = cls.retrieve(screen_name=screen_name)
            permission.roles.append(role)
            permission.update()

        @classmethod
        def forbit(cls, screen_name, role):
            """UserまたはGroupの許可を取りやめます """
            permission = cls.retrieve(screen_name=screen_name)
            permission.roles.remove(role)
            permission.update()

        @classmethod
        def is_accepted(cls, screen_name, node):
            """UserまたはGroupが許可されているか確認します
            """
            perm = cls.retrieve(screen_name=screen_name)
            for role in perm.roles:
                if isinstance(node, User):
                    if node in role.users:
                        return True
                    for group in role.groups:
                        if group.is_in(node):
                            return True
                else:
                    if node in role.groups:
                        return True
            return False

        @classmethod
        def is_forbidden(cls, screen_name, node):
            """UserまたはGroupが許可されていないか確認します
            """
            return not cls.is_accepted(screen_name, node)

    def after_create(target, conn, **kw):
        try:
            DatabaseManager.start_session()
            for perm_ in permissions:
                perm = Permission()
                perm.name = perm_["name"]
                perm.screen_name = perm_["screen_name"]
                db_.session.add(perm)
            db_.session.commit()
        except Exception as err:
            print(err)
            DatabaseManager.rollback_session()
        finally:
            DatabaseManager.close_session()

    event.listen(Permission.__table__, "after_create", after_create)

    return Permission


def inner_permission(db_, permissions):
    from .nodes import User, Group, Node, UserGroup, Role, InnerRole

    inner_role_permission = Table(
        db_.Model.prefix + "_inner_role_permission",
        db_.metadata,
        Column(
            "role_id",
            String(64),
            ForeignKey("mitama_inner_role._id", ondelete="CASCADE")
        ),
        Column(
            "permission_id",
            String(64),
            ForeignKey(
                db_.Model.prefix + "_inner_permission._id",
                ondelete="CASCADE"
            )
        ),
        extend_existing=True
    )

    class InnerPermission(db_.Model):
        name = Column(String(64))
        screen_name = Column(String(64), unique=True)
        roles = relationship(
            "InnerRole",
            secondary=inner_role_permission,
            cascade="all, delete"
        )

        @classmethod
        def accept(cls, screen_name, role):
            """特定のRoleに許可します """
            permission = cls.retrieve(screen_name=screen_name)
            permission.roles.append(role)
            permission.update()

        @classmethod
        def forbit(cls, screen_name, role):
            """UserまたはGroupの許可を取りやめます """
            permission = cls.retrieve(screen_name=screen_name)
            permission.roles.remove(role)
            permission.update()

        @classmethod
        def is_accepted(cls, screen_name, group, user):
            """UserまたはGroupが許可されているか確認します
            """
            if isinstance(group, Node):
                group = group.object
                if isinstance(group, User):
                    return False
            rel = UserGroup.retrieve(user=user, group=group)
            permission = cls.retrieve(screen_name=screen_name)
            for role in permission.roles:
                if rel in role.relations:
                    return True
            return False

        @classmethod
        def is_forbidden(cls, screen_name, group, node):
            """UserまたはGroupが許可されていないか確認します
            """
            return not cls.is_accepted(screen_name, node)

    def after_create(target, conn, **kw):
        try:
            DatabaseManager.start_session()
            for perm_ in permissions:
                perm = InnerPermission()
                perm.name = perm_["name"]
                perm.screen_name = perm_["screen_name"]
                db_.session.add(perm)
            db_.session.commit()
        except Exception as err:
            print(err)
            DatabaseManager.rollback_session()
        finally:
            DatabaseManager.close_session()

    event.listen(InnerPermission.__table__, "after_create", after_create)

    return InnerPermission



import os
from pathlib import Path

from mitama.app import App as BaseApp
from mitama.app import Router
from mitama.app.method import view, post
from mitama.models import Permission, InnerPermission
from mitama.utils.controllers import static_files, mitama_favicon, mitama_service_worker, mitama_manifest
from mitama.utils.middlewares import SessionMiddleware, CsrfMiddleware

from .controller import (
    HomeController,
    UsersController,
    GroupsController,
    RegisterController,
    SessionController,
    SLOController,
    ACSController,
    AppsController,
)
from .middleware import InitializeMiddleware

with open(Path(os.path.dirname(__file__)) / "static/icon.png", "rb") as f:
    icon = f.read()


class App(BaseApp):
    name = "Mitama Portal"
    description = "Mitamaのアプリポータルです。他のアプリを確認できる他、配信の設定やグループの編集、ユーザーの招待ができます。"
    icon = icon

    @property
    def view(self):
        view = super().view
        view.globals.update(
            permission=Permission.is_accepted,
            inner_permission=InnerPermission.is_accepted,
        )
        return view

    @property
    def router(self):
        return Router(
            [
                view("/static/<path:path>", static_files()),
                view("/favicon.ico", mitama_favicon()),
                view("/sw.js", mitama_service_worker()),
                view("/manifest.json", mitama_manifest()),
                Router(
                    [
                        view("/setup", RegisterController, "setup"),
                        view("/signup", RegisterController, "signup"),
                        view("/login", SessionController, "login"),
                        Router(
                            [
                                view("/", HomeController),
                                view("/settings", HomeController, "settings"),
                                view("/logout", SessionController, "logout"),
                                view("/users", UsersController, "list"),
                                view("/users/invite", UsersController, "create"),
                                view("/users/invite/<id>/delete", UsersController, "cancel"),
                                view("/users/<id>", UsersController, "retrieve"),
                                view("/users/<id>/settings", UsersController, "update_profile"),
                                view("/users/<id>/settings/profile", UsersController, "update_profile"),
                                view("/users/<id>/settings/password", UsersController, "update_password"),
                                view("/users/<id>/settings/notification", UsersController, "update_notification"),
                                view("/users/<id>/delete", UsersController, "delete"),
                                view("/groups", GroupsController, "list"),
                                view("/groups/create", GroupsController, "create"),
                                view("/groups/<id>", GroupsController, "retrieve"),
                                post("/groups/<id>/append", GroupsController, "append"),
                                view("/groups/<id>/groups/<gid>/remove", GroupsController, "remove_group"),
                                view("/groups/<id>/users/<uid>/remove", GroupsController, "remove_user"),
                                view(
                                    "/groups/<id>/users/<uid>/accept",
                                    GroupsController,
                                    "accept",
                                ),
                                view(
                                    "/groups/<id>/users/<uid>/forbit",
                                    GroupsController,
                                    "forbit",
                                ),
                                view("/groups/<id>/settings", GroupsController, "update"),
                                view("/groups/<id>/delete", GroupsController, "delete"),
                                view("/apps", AppsController, "list"),
                                view("/apps/settings", AppsController, "update"),
                                view("/acs/post<hoge:re:.*>", ACSController, "post"),
                                view("/acs/redirect<hoge:re:.*>", ACSController, "redirect"),
                                view("/slo/post<hoge:re:.*>", SLOController, "post"),
                                view("/slo/redirect<hoge:re:.*>", SLOController, "redirect"),
                            ],
                            middlewares=[InitializeMiddleware, SessionMiddleware],
                        ),
                    ],
                    middlewares=[CsrfMiddleware]
                )
            ]
        )



import json
import time
import base64
from collections.abc import MutableMapping

from Crypto.Cipher import AES
from Crypto.Random import get_random_bytes
from Crypto.Util import Padding


class Session(MutableMapping):
    def __init__(self, identity, *, data, new, max_age=None):
        self._changed = False
        self._mapping = {}
        self._identity = identity if data != {} else None
        self._new = new if data != {} else True
        self._max_age = max_age
        created = data.get("created", None) if data else None
        session_data = data.get("session", None) if data else None
        now = int(time.time())
        age = now - created if created else now
        if max_age is not None and age > max_age:
            session_data = None
        if self._new or created is None:
            self._created = now
        else:
            self._created = created

        if session_data is not None:
            self._mapping.update(session_data)

    @property
    def new(self):
        return self._new

    @property
    def identity(self):
        return self._identity

    @property
    def created(self):
        return self._created

    @property
    def empty(self):
        return not bool(self._mapping)

    @property
    def max_age(self):
        return self._max_age

    @max_age.setter
    def max_age(self, value):
        self._max_age = value

    def changed(self):
        self._changed = True

    def invalidate(self):
        self._changed = True
        self._mapping = {}

    def set_new_identity(self, identity):
        if not self._new:
            raise RuntimeError("Can't change identity for a session which is not new")
        self._identity = identity

    def __len__(self):
        return len(self._mapping)

    def __iter__(self):
        return iter(self._mapping)

    def __containers__(self, key):
        return key in self._mapping

    def __getitem__(self, key):
        return self._mapping[key]

    def __setitem__(self, key, value):
        self._mapping[key] = value
        self._changed = True

    def __delitem__(self, key):
        del self._mapping[key]
        self._changed = True


class EncryptedCookieStorage:
    def __init__(
        self,
        secret_key,
        *,
        cookie_name="MITAMA_SESSION",
        domain=None,
        max_age=None,
        path="/",
        secure=None,
        httponly=True,
        encoder=json.dumps,
        decoder=json.loads
    ):
        self._cookie_name = cookie_name
        self._cookie_params = dict(
            domain=domain, max_age=max_age, path=path, secure=secure, httponly=httponly
        )
        self._max_age = max_age
        self._encoder = encoder
        self._decoder = decoder
        if isinstance(secret_key, str):
            secret_key = secret_key.encode("utf-8")
        elif isinstance(secret_key, (bytes, bytearray)):
            pass
        self.secret_key = secret_key

    @property
    def cookie_name(self):
        return self._cookie_name

    @property
    def max_age(self):
        return self._max_age

    @property
    def cookie_params(self):
        return self._cookie_params

    def _get_session_data(self, session):
        if not session.empty:
            data = {"created": session.created, "session": session._mapping}
        else:
            data = {}
        return data

    def load_session(self, request):
        cookie = self.load_cookie(request)
        if cookie is None:
            return Session(None, data=None, new=True, max_age=self.max_age)
        else:
            try:
                data = base64.urlsafe_b64decode(cookie.encode("utf-8"))
                iv = data[:16]
                cipher = AES.new(self.secret_key, AES.MODE_CBC, iv)
                decrypted = cipher.decrypt(
                    data[16:]
                )
                without_padding = Padding.unpad(
                    decrypted,
                    16,
                    "pkcs7"
                )
                data = self._decoder(
                    without_padding.decode("utf-8")
                )
                return Session(None, data=data, new=False, max_age=self.max_age)
            except (TypeError, ValueError) as err:
                print(err)
                return Session(None, data=None, new=True, max_age=self.max_age)

    def save_session(self, request, response, session):
        if session.empty:
            return self.save_cookie(response, "", max_age=session.max_age)
        iv = get_random_bytes(16)
        cookie_data = self._encoder(self._get_session_data(session)).encode("utf-8")
        cipher = AES.new(self.secret_key, AES.MODE_CBC, iv)
        with_padding = Padding.pad(
            cookie_data,
            16,
            "pkcs7"
        )
        encrypted = base64.urlsafe_b64encode(iv + cipher.encrypt(with_padding)).decode("utf-8")
        self.save_cookie(
            response,
            encrypted,
            max_age=session.max_age,
        )

    def load_cookie(self, request):
        cookie = request.cookies.get(self._cookie_name)
        return cookie

    def save_cookie(self, response, cookie_data, *, max_age=None):
        params = dict(self._cookie_params)
        params_ = dict()
        for k in params.keys():
            params_[k.replace("_", "-")] = params[k]
        if max_age is not None:
            params_["max-age"] = max_age
            params_["expires"] = time.strftime(
                "%a, %d-%b-%Y %T GMT", time.gmtime(time.time() + max_age)
            )
        if not cookie_data:
            response.del_cookie(
                self._cookie_name, domain=params_["domain"], path=params_["path"]
            )
        else:
            response.set_cookie(self._cookie_name, cookie_data, **params_)



class Event:
    def __init__(self, doc=None):
        self.__doc__ = doc

    def __get__(self, obj, objtype=None):
        if obj is None:
            return Self
        return EventHandler(self, obj)

    def __set__(self, obj, objtype=None):
        pass

class EventHandler:
    def __init__(self, event, obj):
        self.event = event
        self.obj = obj
    def _getfunctionlist(self):
        try:
            eventhandler = self.obj.__eventhandler__
        except AttributeError:
            eventhandler = self.obj.__eventhandler__ = {}
        return eventhandler.setdefault(self.event, [])
    def __iadd__(self, func):
        self._getfunctionlist().append(func)
        return self
    def __isub__(self, func):
        self._getfunctionlist().remove(func)
        return self
    def __call__(self,earg=None):
        for func in self._getfunctionlist():
            func(self.obj, earg)




import re


class _Route:
    def __init__(self, methods, path, handler, method_name):
        self.methods = methods
        self.path = _Path(path)
        self.handler = handler
        self.method_name = method_name
        pass

    def match(self, request):
        method = request.method
        path = request.subpath if hasattr(request, "subpath") else request.path
        args = self.path.match(path)
        if method in self.methods and args != False:
            request.params = args
            return request, self.handler, self.method_name
        else:
            return False


class _GroupRoute:
    def __init__(self, path, router):
        self.path = path
        self.router = router
        pass

    def match(self, request):
        path = request.subpath if hasattr(request, "subpath") else request.path
        if str(self.path) != "" and str(self.path) != "/":
            if path[: len(self.path)] != self.path:
                return False
            else:
                path = path[len(self.path) :]
                request.subpath = path
        return self.router.match(request)


def _re_flatten(p):
    if "(" not in p:
        return p
    return re.sub(
        r"(\\*)(\(\?P<[^>]+>|\((?!\?))",
        lambda m: m.group(0) if len(m.group(1)) % 2 else m.group(1) + "(?:",
        p,
    )


class _Path:
    rule_syntax = re.compile(
        "(\\\\*)(?:(?:<([a-zA-Z_][a-zA-Z_0-9]*)?(?::([a-zA-Z_]*)(?::((?:\\\\.|[^\\\\>])+)?)?)?>))"
    )
    filters = {
        "re": lambda conf: (_re_flatten(conf or "[^/]+"), None, None),
        "int": lambda conf: (r"-?\d+", int, lambda x: str(int(x))),
        "float": lambda conf: (r"-?[\d.]+", float, lambda x: str(float(x))),
        "path": lambda conf: (r".+?", None, None),
    }
    default_filter = "re"

    def __init__(self, path):
        self.raw = path
        self.builder = []
        anons = 0
        pattern = ""
        keys = []
        filters = []
        is_static = True
        for key, mode, conf in self._itertoken(path):
            if mode:
                is_static = False
                if mode == "default":
                    mode = self.default_filter
                mask, in_filter, out_filter = self.filters[mode](conf)
                if not key:
                    pattern += "(?:%s)" % mask
                    key = "anon%d" % anons
                    anons += 1
                else:
                    pattern += "(?P<%s>%s)" % (key, mask)
                    keys.append(key)
                if in_filter:
                    filters.append((key, in_filter))
            elif key:
                pattern += re.escape(key)

        try:
            re_pattern = re.compile("^(%s)$" % pattern)
            re_match = re_pattern.match
        except re.error as e:
            raise _RouteSyntaxError("")

        if filters:

            def getargs(path):
                url_args = re_match(path).groupdict()
                for name, wildcard_filter in filters:
                    try:
                        url_args[name] = wildcard_filter(url_args[name])
                    except ValueError:
                        raise RoutingError()
                return url_args

        elif re_pattern.groupindex:

            def getargs(path):
                return re_match(path).groupdict()

        else:
            getargs = None
        flatpat = _re_flatten(pattern)
        self.rule = (path, re.compile("(^%s$)" % flatpat), getargs)

    def _itertoken(self, path):
        offset = 0
        prefix = ""
        for match in self.rule_syntax.finditer(path):
            prefix += path[offset : match.start()]
            g = match.groups()
            if len(g[0]) % 2:
                prefix += match.group(0)[len(g[0]) :]
                offset = match.end()
                continue
            if prefix:
                yield prefix, None, None
            name, filtr, conf = g[1:4]
            yield name, filtr or "default", conf or None
            offset, prefix = match.end(), ""
        if offset <= len(path) or prefix:
            yield prefix + path[offset:], None, None

    def match(self, target):
        path, flatpat, getargs = self.rule
        if flatpat.match(target) != None:
            return getargs(target) if getargs else {}
        else:
            return False


def view(path, handler, method=None):
    """GET, POSTのルーティング先を指定

    ブラウザで見れる一般的なルーティング先を作成します。
    :param path: マッチするパス
    :param handler: リクエストハンドラ
    :return: _Routeインスタンス
    """
    return _Route(["GET", "POST"], path, handler, method)


def any(path, handler, method=None):
    """メソッドを考慮しないルーティング先を指定

    メソッドに関係なくマッチするルーティング先を作成します。
    :param path: マッチするパス
    :param handler: リクエストハンドラ
    :return: _Routeインスタンス
    """
    return _Route(
        ["GET", "POST", "PATCH", "PUT", "DELETE", "OPTION", "HEAD"],
        path,
        handler,
        method,
    )


def post(path, handler, method=None):
    """POSTのルーティング先を指定

    POSTメソッドのルーティング先を作成します。
    :param path: マッチするパス
    :param handler: リクエストハンドラ
    :return: _Routeインスタンス
    """
    return _Route(["POST"], path, handler, method)


def patch(path, handler, method=None):
    """PATCHのルーティング先を指定

    PATCHメソッドのルーティング先を作成します。
    :param path: マッチするパス
    :param handler: リクエストハンドラ
    :return: _Routeインスタンス
    """
    return _Route(["PATCH"], path, handler, method)


def head(path, handler, method=None):
    """HEADのルーティング先を指定

    HEADメソッドのルーティング先を作成します。
    :param path: マッチするパス
    :param handler: リクエストハンドラ
    :return: _Routeインスタンス
    """
    return _Route(["HEAD"], path, handler, method)


def option(path, handler, method=None):
    """OPTIONのルーティング先を指定

    OPTIONメソッドのルーティング先を作成します。
    :param path: マッチするパス
    :param handler: リクエストハンドラ
    :return: _Routeインスタンス
    """
    return _Route(["OPTION"], path, handler, method)


def put(path, handler, method=None):
    """PUTのルーティング先を指定

    PUTメソッドのルーティング先を作成します。
    :param path: マッチするパス
    :param handler: リクエストハンドラ
    :return: _Routeインスタンス
    """
    return _Route(["PUT"], path, handler, method)


def get(path, handler, method=None):
    """GETのルーティング先を指定

    GETメソッドのルーティング先を作成します。
    :param path: マッチするパス
    :param handler: リクエストハンドラ
    :return: _Routeインスタンス
    """
    return _Route(["GET"], path, handler, method)


def delete(path, handler, method=None):
    """DELETEのルーティング先を指定

    DELETEメソッドのルーティング先を作成します。
    :param path: マッチするパス
    :param handler: リクエストハンドラ
    :return: _Routeインスタンス
    """
    return _Route(["DELETE"], path, handler, method)


def group(path, router):
    """グループルーティング先を指定

    前方一致でルーティング先に中継するルーティング先を作成します。
    :param path: マッチするパス
    :param router: 渡すルーティングエンジン
    :return: _GroupRouteインスタンス
    """
    return _GroupRoute(path, router)



from .fields import *
from .errors import *

class Form:
    def __new__(cls, data=None):
        self = object.__new__(cls)
        ignored_attrs = ['fields']
        pairs = [
            (name, getattr(self, name).instance())
            for name in dir(self)
            if name not in ignored_attrs and isinstance(getattr(self, name), Field)
        ]
        self.fields = dict(pairs)
        for key in self.fields.keys():
            if self.fields[key].name == None:
                self.fields[key].name = key
        return self

    def __init__(self, data=None):
        for field in self.fields.values():
            if isinstance(field, DictFieldInstance):
                field.data = {}
                for k in data.keys():
                    if '.' in k:
                        k_ = k.split('.')
                        name = k_[0]
                        key = '.'.join(k_[1:])
                        if name == field.name:
                            if field.listed:
                                field.data[key] = data.getlist(k)
                            else:
                                field.data[key] = data.get(k)
            elif field.listed:
                field.data = data.getlist(field.name)
            elif field.name in data:
                field.data = data.get(field.name)
            field.validate()

    def __getitem__(self, key):
        return self.fields[key].data

    def reset(self):
        for value in self.fields.values():
            value.reset()



from mitama.app import App, Router
from mitama.utils.controllers import static_files
from mitama.app.method import view

from .controller import WelcomeController


class App(App):
    # name = 'MyApp'
    # description = 'This is my App.'
    router = Router(
        [
            view("/", WelcomeController),
            view("/static/<path:path>", static_files()),
        ]
    )



import inspect
import os


class Builder(object):
    """Appにメタ情報を入力して生成するビルダー

    プロジェクトディレクトリのパスやインストール先、プロジェクト内のアプリ用ディレクトリのパスをAppに設定し、インスタンスを返却します。
    特にこれを弄るケースは想定していませんが、独自の挙動を付けたかったら継承して作っても良いかもしれません。
    アプリのパッケージ直下のAppBuilderが起動されるので、:file:`__init__.py` に :samp:`class AppBuilder(Builder)` を定義してください。
    """

    app = None

    def __init__(self):
        self.data = {}
        pass

    def set_path(self, path):
        self.data["path"] = path

    def set_screen_name(self, screen_name):
        self.data["screen_name"] = screen_name

    def set_package(self, package):
        self.data["package"] = package

    def set_project_dir(self, path):
        self.data["project_dir"] = path

    @property
    def path(self):
        return self.data.get("path")

    @property
    def screen_name(self):
        return self.data.get("screen_name")

    @property
    def package(self):
        return self.data.get("package")

    @property
    def project_dir(self):
        return self.data.get("project_dir")

    @property
    def project_root_dir(self):
        return self.data.get("project_root_dir")

    def build(self):
        install_dir = os.path.dirname(inspect.getfile(self.__class__))
        self.data["install_dir"] = install_dir
        return self.app(**self.data)




from mitama.app import Controller
from mitama.app.http import Response


class WelcomeController(Controller):
    def handle(self, request):
        template = self.view.get_template("welcome.html")
        return Response.render(template)



#!/usr/bin/python
import os
from base64 import b64encode
from pathlib import Path

import magic
import markdown
from jinja2 import Markup, Environment, ChoiceLoader, FileSystemLoader
import uuid

from mitama.noimage import load_noimage_app

from .http import Request, Response


def dataurl(blob):
    f = magic.Magic(mime=True, uncompress=True)
    mime = f.from_buffer(blob)
    return "data:" + mime + ";base64," + b64encode(blob).decode()


class App:
    template_dir = "templates"
    description = ""
    name = ""
    models = list()

    @property
    def icon(self):
        return load_noimage_app()

    def __init__(self, screen_name, path, package, project_dir=None, install_dir=os.path.dirname(__file__), **kwargs):
        self.screen_name = screen_name
        self.path = path
        self.package = package
        self.project_dir = Path(project_dir)
        self.install_dir = Path(install_dir)
        self.params = kwargs
        self.router._app = self
        self.init_app()

    def init_app(self):
        pass

    def wsgi(self, env, start_response):
        request = Request(env)
        try:
            response = self(request)
            body = response.start(request, start_response)
        except Exception as err:
            print(err)
            body = self.error(request, 500).start(request, start_response)
        return body

    def __call__(self, request):
        if not isinstance(request, Request):
            request = Request.from_request(request)
        result = self.router.match(request)
        if result:
            request, handle, method = result
            return handle(request)
        else:
            return self.error(request, 404)

    def set_middleware(self, middlewares):
        self.app.middlewares.extend(middlewares)

    def convert_fullurl(self, req, url, scheme=None):
        scheme = scheme if scheme is not None else req.scheme
        hostname = req.host
        path = self.path
        if path[0] != "/":
            path = "/" + path
        if path[-1] == "/":
            path = path[0:-2]
        if url[0] != "/":
            url = "/" + url
        return scheme + "://" + hostname + path + url

    def convert_url(self, url):
        path = self.path
        if path[0] != "/":
            path = "/" + path
        if path[-1] == "/":
            path = path[0:-2]
        if url[0] != "/":
            url = "/" + url
        return path + url

    def revert_url(self, url):
        path = self.path
        if path[0] != "/":
            path = "/" + path
        if path[-1] == "/":
            path = path[0:-2]
        url = str(url)
        url = url[len(path) :]
        return url

    @property
    def view(self):
        toolkit = Path(os.path.dirname(__file__)) / "templates"
        self._view = Environment(
            loader=ChoiceLoader([
                FileSystemLoader(self.install_dir / self.template_dir),
                FileSystemLoader(toolkit),
            ])
        )

        def filter_user(arg):
            return [user for user in arg if user.__class__.__name__ == "User"]

        def filter_group(arg):
            return [group for group in arg if group.__class__.__name__ == "Group"]

        def markdown_(text):
            return Markup(markdown.markdown(text, extensions=['fenced_code', 'def_list', 'attr_list', 'abbr', 'footnotes', 'tables', 'wikilinks', 'toc', 'nl2br']))

        self._view.filters["user"] = filter_user
        self._view.filters["group"] = filter_group
        self._view.filters["markdown"] = markdown_
        self._view.globals.update(
            url=self.convert_url,
            fullurl=self.convert_fullurl,
            dataurl=dataurl,
            uuid=uuid.uuid4
        )
        return self._view

    def error(self, request, code):
        template = self.view.get_template(str(code) + ".html")
        return Response.render(template, status=code)

    def match(self, request):
        result = self.router.match(request)
        if result == False:
            return False
        else:
            request, handle, method = result

            def _handle(request):
                request.app = self
                return handle(request)

            return request, _handle, method

    def save_params(self):
        self.params

    def uninstall(self):
        db = DatabaseManager()
        for model in self.models:
            model.__table__.drop(db.engine)

    def model(self, modelname):
        for model in self.models:
            if model.__class__.__name__ == modelname:
                return model

def _session_middleware():
    import base64

    from Crypto.Random import get_random_bytes

    from mitama.app import Middleware
    from mitama.app.http.session import EncryptedCookieStorage

    if "MITAMA_SESSION_KEY" in os.environ:
        session_key = os.environ["MITAMA_SESSION_KEY"]
    elif os.path.exists(".tmp/MITAMA_SESSION_KEY"):
        with open(".tmp/MITAMA_SESSION_KEY", "r") as f:
            session_key = f.read()
    else:
        key = get_random_bytes(16)
        session_key = base64.urlsafe_b64encode(key).decode("utf-8")
        if not os.path.exists(".tmp"):
            os.mkdir(".tmp")
        with open(".tmp/MITAMA_SESSION_KEY", "w") as f:
            f.write(session_key)

    class SessionMiddleware(Middleware):
        fernet_key = session_key

        def __init__(self):
            secret_key = base64.urlsafe_b64decode(self.fernet_key.encode("utf-8"))
            cookie_storage = EncryptedCookieStorage(secret_key)
            self.storage = cookie_storage

        def process(self, request, handler):
            request["mitama_session_storage"] = self.storage
            raise_response = False
            response = handler(request)
            if not isinstance(response, Response):
                return response
            session = request.get("mitama_session")
            if session is not None:
                if session._changed:
                    self.storage.save_session(request, response, session)
            if raise_response:
                raise response
            return response

    return SessionMiddleware




import base64
import hashlib
import random
import secrets
from pywebpush import webpush

import bcrypt
import jwt
import magic
import json
from Crypto.Random import get_random_bytes
from Crypto.Hash import SHA256

from mitama.db import ForeignKey, relationship, Table, backref
from mitama.db.types import Column, LargeBinary
from mitama.db.types import String
from mitama.db.model import UUID
from mitama.noimage import load_noimage_group, load_noimage_user
from mitama._extra import _classproperty

from .core_db import db

secret = secrets.token_hex(32)


class AuthorizationError(Exception):
    INVALID_TOKEN = 0
    WRONG_PASSWORD = 1
    USER_NOT_FOUND = 2

    def __init__(self, code):
        self.code = code

    @property
    def message(self):
        return [
            "トークンが不正です",
            "パスワードが間違っています",
            "ユーザーが見つかりません"
        ][self.code]

    def __str__(self):
        return self.message


user_group = Table(
    "mitama_user_group",
    db.metadata,
    Column("_id", String(64), default=UUID(), primary_key=True),
    Column(
        "group_id",
        String(64),
        ForeignKey("mitama_group._id", ondelete="CASCADE")
    ),
    Column(
        "user_id",
        String(64),
        ForeignKey("mitama_user._id", ondelete="CASCADE")
    ),
)


class UserGroup(db.Model):
    __table__ = user_group
    _id = user_group.c._id
    group_id = user_group.c.group_id
    user_id = user_group.c.user_id
    user = relationship("User")
    group = relationship("Group")


class AbstractNode(object):
    _icon = Column(LargeBinary)
    _name = Column("name", String(255))
    _screen_name = Column("screen_name", String(255))
    _name_proxy = list()
    _screen_name_proxy = list()
    _icon_proxy = list()

    @property
    def name(self):
        name = self._name
        for fn in self._name_proxy:
            name = fn(name)
        return name

    @property
    def screen_name(self):
        screen_name = self._screen_name
        for fn in self._screen_name_proxy:
            screen_name = fn(screen_name)
        return screen_name

    def to_dict(self):
        return {
            "_id": self._id,
            "name": self.name,
            "screen_name": self.screen_name,
        }

    @property
    def icon(self):
        if self._icon is not None:
            icon = self._icon
        else:
            icon = self.load_noimage()
        for fn in self._icon_proxy:
            icon = fn(icon)
        return icon

    @name.setter
    def name(self, value):
        self._name = value

    @screen_name.setter
    def screen_name(self, value):
        self._screen_name = value

    @icon.setter
    def icon(self, value):
        self._icon = value

    def icon_to_dataurl(self):
        f = magic.Magic(mime=True, uncompress=True)
        mime = f.from_buffer(self.icon)
        return "".join([
            "data:",
            mime,
            ";base64,",
            base64.b64encode(self.icon).decode()
        ])

    @classmethod
    def add_name_proxy(cls, fn):
        cls._name_proxy.append(fn)

    @classmethod
    def add_screen_name_proxy(cls, fn):
        cls._screen_name_proxy.append(fn)

    @classmethod
    def add_icon_proxy(cls, fn):
        cls._icon_proxy.append(fn)

    @classmethod
    def retrieve(cls, _id=None, screen_name=None, **kwargs):
        if _id is not None:
            return super().retrieve(_id=_id)
        elif screen_name is not None:
            return super().retrieve(_screen_name=screen_name)
        else:
            return super().retrieve(**kwargs)

    def __eq__(self, other):
        return self._id == other._id


class User(AbstractNode, db.Model):
    """ユーザーのモデルクラスです

    :param _id: 固有のID
    :param screen_name: ログイン名
    :param name: 名前
    :param password: パスワード
    :param icon: アイコン
    """

    __tablename__ = "mitama_user"
    _id = Column(
        String(64),
        default=UUID("user"),
        primary_key=True,
        nullable=False
    )
    _project = None
    _token = Column(String(64))
    email = Column(String(64), nullable=False)
    password = Column(String(255))
    groups = relationship(
        "Group",
        secondary=user_group,
    )

    def to_dict(self, only_profile=False):
        profile = super().to_dict()
        if not only_profile:
            profile["groups"] = [p.to_dict(True) for p in self.groups()]
        return profile

    def load_noimage(self):
        return load_noimage_user()

    def password_check(self, password):
        password = base64.b64encode(
            hashlib.sha256(password.encode() * 10).digest()
        )
        password_ = self.password
        if isinstance(password_, str):
            password_ = password_.encode()
        return bcrypt.checkpw(password, password_)

    @classmethod
    def password_auth(cls, screen_name, password):
        """ログイン名とパスワードで認証します

        :param screen_name: ログイン名
        :param password: パスワード
        :return: Userインスタンス
        """
        try:
            user = cls.retrieve(_screen_name=screen_name)
            if user is None:
                raise AuthorizationError(AuthorizationError.USER_NOT_FOUND)
        except Exception:
            raise AuthorizationError(AuthorizationError.USER_NOT_FOUND)
        password = base64.b64encode(
            hashlib.sha256(password.encode() * 10).digest()
        )
        password_ = user.password
        if isinstance(password_, str):
            password_ = password_.encode()
        if bcrypt.checkpw(password, password_):
            return user
        else:
            raise AuthorizationError(AuthorizationError.WRONG_PASSWORD)

    def valid_password(self, password):
        """パスワードが安全か検証します

        :param password: パスワードのプレーンテキスト
        :return: 検証済みパスワード
        """
        if self._project is None:
            return password
        project = self._project
        if project.password_validation is None:
            return password
        MIN_PASSWORD_LEN = project.password_validation.get(
            'min_password_len',
            None
        )
        COMPLICATED_PASSWORD = project.password_validation.get(
            'complicated_password',
            False
        )

        if MIN_PASSWORD_LEN and len(password) < MIN_PASSWORD_LEN:
            raise ValueError('パスワードは{}文字以上である必要があります'.format(MIN_PASSWORD_LEN))

        if (
            COMPLICATED_PASSWORD and (
                not any(c.isdigit() for c in password)
            ) or (not any(c.isalpha() for c in password))
        ):
            raise ValueError('パスワードは数字とアルファベットの両方を含む必要があります')

        return password

    def set_password(self, password):
        """パスワードをハッシュ化します

        :param password: パスワードのプレーンテキスト
        :return: パスワードハッシュ
        """
        password = self.valid_password(password)
        salt = bcrypt.gensalt()
        password = base64.b64encode(
            hashlib.sha256(password.encode() * 10).digest()
        )
        self.password = bcrypt.hashpw(password, salt)

    def mail(self, subject, body, type="html"):
        self._project.send_mail(self.email, subject, body, type)

    def get_jwt(self):
        nonce = "".join([str(random.randint(0, 9)) for i in range(16)])
        result = jwt.encode({
            "id": self._id,
            "nonce": nonce
        }, secret, algorithm="HS256")
        return result

    @classmethod
    def check_jwt(cls, token):
        """JWTからUserインスタンスを取得します

        :param token: JWT
        :return: Userインスタンス
        """
        try:
            result = jwt.decode(token, secret, algorithms="HS256")
        except jwt.exceptions.InvalidTokenError:
            raise AuthorizationError(AuthorizationError.INVALID_TOKEN)
        return cls.retrieve(result["id"])

    def is_ancestor(self, node):
        if not isinstance(node, Group) and not isinstance(node, User):
            raise TypeError("Checking object must be Group or User instance")
        layer = self.groups
        while len(layer) > 0:
            if isinstance(node, Group) and node in layer:
                return True
            else:
                for node_ in layer:
                    if isinstance(node, User) and node_.is_in(node):
                        return True
            layer_ = list()
            for node_ in layer:
                layer_.extend(node_.groups)
            layer = layer_
        return False

    def push(self, data):
        for subscription in self.subscriptions:
            subscription.push(data)


class Group(AbstractNode, db.Model):
    """グループのモデルクラスです

    :param _id: 固有のID
    :param screen_name: ドメイン名
    :param name: 名前
    :param icon: アイコン
    """

    __tablename__ = "mitama_group"
    _id = Column(
        String(64),
        default=UUID("group"),
        primary_key=True,
        nullable=False
    )
    _project = None
    users = relationship(
        "User",
        secondary=user_group,
    )
    parent_id = Column(String(64), ForeignKey("mitama_group._id"))
    groups = relationship(
        "Group",
        backref=backref("parent", remote_side=[_id]),
    )

    def to_dict(self, only_profile=False):
        profile = super().to_dict()
        if not only_profile:
            profile["parent"] = self.parent.to_dict()
            profile["groups"] = [n.to_dict(True) for n in self.groups]
            profile["users"] = [n.to_dict(True) for n in self.users]
        return profile

    def load_noimage(self):
        return load_noimage_group()

    @_classproperty
    def relation(cls):
        return Column(String, ForeignKey("mitama_group._id"), nullable=False)

    @_classproperty
    def relations(cls):
        return relationship("mitama_group._id", cascade="all, delete")

    @_classproperty
    def relation_or_null(cls):
        return Column(String, ForeignKey("mitama_group._id"), nullable=True)

    @classmethod
    def tree(cls):
        return Group.query.filter(Group.parent is None).all()

    def append(self, node):
        if isinstance(node, User):
            self.users.append(node)
        elif isinstance(node, Group):
            self.groups.append(node)
        else:
            raise TypeError("Appending object must be Group or User instance")
        self.query.session.commit()

    def append_all(self, nodes):
        for node in nodes:
            if isinstance(node, User):
                self.users.append(node)
            elif isinstance(node, Group):
                self.groups.append(node)
            else:
                raise TypeError(
                    "Appending object must be Group or User instance"
                )
        self.query.session.commit()

    def remove(self, node):
        if not isinstance(node, Group) and not isinstance(node, User):
            raise TypeError("Removing object must be Group or User instance")
        if isinstance(node, Group):
            self.groups.remove(node)
        else:
            self.users.remove(node)
        self.query.session.commit()

    def remove_all(self, nodes):
        for node in nodes:
            if not isinstance(node, Group) and not isinstance(node, User):
                raise TypeError(
                    "Appending object must be Group or User instance"
                )
            if isinstance(node, Group):
                self.groups.remove(node)
            else:
                self.users.remove(node)
        self.query.session.commit()

    def is_ancestor(self, node):
        if not isinstance(node, Group) and not isinstance(node, User):
            raise TypeError("Checking object must be Group or User instance")
        if self.parent is None:
            return False
        layer = [self.parent]
        while len(layer) > 0:
            if isinstance(node, Group) and node in layer:
                return True
            else:
                for node_ in layer:
                    if isinstance(node, User) and node_.is_in(node):
                        return True
            layer_ = list()
            for node_ in layer:
                if node_.parent is not None:
                    layer_.extend([node_.parent])
            layer = layer_
        return False

    def is_descendant(self, node):
        if not isinstance(node, Group) and not isinstance(node, User):
            raise TypeError("Checking object must be Group or User instance")
        layer = self.groups
        while len(layer) > 0:
            if node in layer:
                return True
            layer_ = list()
            for node_ in layer:
                layer_.extend(node_.groups)
            layer = layer_
        return False

    def is_in(self, node):
        if isinstance(node, User):
            return node in self.users
        elif isinstance(node, Group):
            return node in self.groups
        else:
            raise TypeError("Checking object must be Group or User instance")

    def __contains__(self, node):
        return self.is_in(node)

    def delete(self):
        """グループを削除します"""
        from mitama.app.hook import HookRegistry
        hook_registry = HookRegistry()
        hook_registry.delete_group(self)
        super().delete()

    def update(self):
        """グループの情報を更新します"""
        super().update()
        from mitama.app.hook import HookRegistry
        hook_registry = HookRegistry()
        hook_registry.update_group(self)

    def create(self):
        """グループを作成します"""
        super().create()
        from mitama.app.hook import HookRegistry
        hook_registry = HookRegistry()
        hook_registry.create_group(self)

    def mail(self, subject, body, type="html", to_all=False):
        for user in self.users:
            user.mail(subject, body, type)
        if to_all:
            for group in self.groups:
                group.mail(subject, body, type, to_all)


def get_random_token():
    s = get_random_bytes(32)
    h = SHA256.new()
    h.update(s)
    return h.hexdigest()


class UserInvite(db.Model):
    __tablename__ = "mitama_user_invite"
    token = Column(String(64), default=get_random_token, unique=True)
    email = Column(String(255))
    screen_name = Column(String(255))
    name = Column(String(255))
    _icon = Column(LargeBinary)
    roles = Column(String(255), default="")

    def load_noimage(self):
        return load_noimage_user()

    @property
    def icon(self):
        return self._icon or self.load_noimage()

    def icon_to_dataurl(self):
        f = magic.Magic(mime=True, uncompress=True)
        mime = f.from_buffer(self.icon)
        return ''.join([
            "data:",
            mime,
            ";base64,",
            base64.b64encode(self.icon).decode()
        ])

    def mail(self, subject, body, type="html"):
        self._project.send_mail(self.email, subject, body, type)


role_user = Table(
    "mitama_role_user",
    db.metadata,
    Column(
        "role_id",
        String(64),
        ForeignKey("mitama_role._id", ondelete="CASCADE")
    ),
    Column(
        "user_id",
        String(64),
        ForeignKey("mitama_user._id", ondelete="CASCADE")
    )
)

role_group = Table(
    "mitama_role_group",
    db.metadata,
    Column(
        "role_id",
        String(64),
        ForeignKey("mitama_role._id", ondelete="CASCADE")
    ),
    Column(
        "group_id",
        String(64),
        ForeignKey("mitama_group._id", ondelete="CASCADE")
    )
)


class Role(db.Model):
    __tablename__ = "mitama_role"
    screen_name = Column(String(64), unique=True, nullable=False)
    name = Column(String(64))
    users = relationship(
        "User",
        secondary=role_user,
        backref="roles",
        cascade="all, delete"
    )
    groups = relationship(
        "Group",
        secondary=role_group,
        backref="roles",
        cascade="all, delete"
    )

    def append(self, node):
        if isinstance(node, Group):
            self.groups.append(node)
        else:
            self.users.append(node)
        self.update()

    def remove(self, node):
        if isinstance(node, Group):
            self.groups.remove(node)
        else:
            self.users.remove(node)
        self.update()


role_relation = Table(
    "mitama_role_relation",
    db.metadata,
    Column("_id", String(64), default=UUID(), primary_key=True),
    Column(
        "role_id",
        String(64),
        ForeignKey("mitama_inner_role._id", ondelete="CASCADE")
    ),
    Column(
        "relation_id",
        String(64),
        ForeignKey("mitama_user_group._id", ondelete="CASCADE")
    )
)


class RoleRelation(db.Model):
    __table__ = role_relation
    _id = role_relation.c._id
    role_id = role_relation.c.role_id
    relation_id = role_relation.c.relation_id


class InnerRole(db.Model):
    __tablename__ = "mitama_inner_role"
    screen_name = Column(String(64), unique=True, nullable=False)
    name = Column(String(64))
    relations = relationship(
        "UserGroup",
        secondary=role_relation,
        backref="roles",
        cascade="all, delete"
    )

    def append(self, group, user):
        relation = UserGroup.retrieve(group=group, user=user)
        self.relations.append(relation)
        self.update()

    def remove(self, group, user):
        relation = UserGroup.retrieve(group=group, user=user)
        self.relations.remove(relation)
        self.update()

    def exists(self, group, user):
        relation = UserGroup.retrieve(group=group, user=user)
        return relation in self.relations


class Node(db.Model):
    __tablename__ = "mitama_node"
    user_id = Column(
        String(64),
        ForeignKey("mitama_user._id", ondelete="CASCADE"),
        unique=True
    )
    group_id = Column(
        String(64),
        ForeignKey("mitama_group._id", ondelete="CASCADE"),
        unique=True
    )
    user = relationship(User)
    group = relationship(Group)

    @classmethod
    def retrieve(cls, obj=None, id=None, **kwargs):
        if obj is not None:
            if isinstance(obj, str):
                if obj.split("-")[0] == "user":
                    obj = User.retrieve(obj)
                else:
                    obj = Group.retrieve(obj)
            try:
                if isinstance(obj, User):
                    return cls.query.filter(cls.user == obj).one()
                else:
                    return cls.query.filter(cls.group == obj).one()
            except Exception:
                if isinstance(obj, User):
                    node = cls()
                    node.user = obj
                    node.create()
                    return node
                else:
                    node = cls()
                    node.group = obj
                    node.create()
                    return node
        else:
            return super().retrieve(id=id, **kwargs)

    @property
    def object(self):
        return self.user if self.user is not None else self.group

    def is_ancestor(self, node):
        if self.group is not None:
            return self.group.is_ancestor(node)
        else:
            return False

    def is_descendant(self, node):
        if self.group is not None:
            return self.group.is_descendant(node)
        else:
            return False

    def is_in(self, node):
        if self.group is not None:
            return self.group.is_in(node)
        else:
            return False

    def __contains__(self, node):
        return self.is_in(node)


class PushSubscription(db.Model):
    __tablename__ = "mitama_push_subscription"
    _project = None
    user_id = Column(
        String(64),
        ForeignKey("mitama_user._id", ondelete="CASCADE")
    )
    user = relationship("User", backref="subscriptions")
    subscription = Column(String(1024))

    def push(self, data):
        try:
            webpush(
                subscription_info=json.loads(self.subscription),
                data=json.dumps(data),
                vapid_private_key=self._project.vapid['private_key'],
                vapid_claims={
                    "sub": "mailto:{}".format(self._project.vapid['mailto'])
                }
            )
        except Exception as err:
            print(err)



import os
import sys
import importlib
import subprocess
import mitama
import inspect
import smtplib
import argparse
import uwsgi
from traceback import print_exc
from pathlib import Path, PosixPath
from email.mime.text import MIMEText
import json
from mitama.app.http import Request
from mitama.app import App, AppRegistry
from mitama.app.app import _session_middleware
from mitama.db import create_engine, DatabaseManager

from . import commands

class Project(App):
    def __init__(
        self,
        *app_builders,
        port = 80,
        password_validation = None,
        project_dir = Path(os.getcwd()),
        mail = {
            "host": "localhost",
            "port": 25,
            "address": "mitama@example.com"
        },
        vapid = {
            "private_key": None,
            "public_key": None,
            "mailto": None
        },
        database = {
            "type": "sqlite"
        },
        **kwargs
    ):
        if not isinstance(project_dir, PosixPath):
            project_dir = Path(project_dir)
        self.install_dir = Path(os.path.dirname(os.path.abspath(__file__)))
        self.project_dir = project_dir

        from mitama.models import User, Group, UserInvite, PushSubscription
        User._project = self
        Group._project = self
        UserInvite._project = self
        PushSubscription._project = self
        self.port = port
        self.mail = mail
        self.vapid = vapid
        self.password_validation = password_validation
        self.apps = AppRegistry()
        self.apps.project = self
        for builder in app_builders:
            if builder.screen_name is None:
                builder.set_screen_name(builder.package)
            builder.set_project_dir(self.project_dir / builder.screen_name)
            app = builder.build()
            self.apps[app.screen_name] = app
        self.config = kwargs

        self.router = self.apps.router()

    def send_mail(self, to, subject, body, type="html"):
        mail = self.mail
        smtp = smtplib.SMTP(mail["host"], mail["port"])
        if smtp.has_extn("STARTTLS"):
            smtp.starttls()
        if "user" in mail:
            smtp.login(mail["user"]["login"], mail["user"]["password"])
        msg = MIMEText(body, type)
        msg["Subject"] = subject
        msg["To"] = to
        msg["From"] = mail["address"]
        smtp.send_message(msg)
        smtp.quit()

    @property
    def arg_parser(self):
        if not hasattr(self, "_arg_parser"):
            self._arg_parser = argparse.ArgumentParser(description="")
            subparser = self._arg_parser.add_subparsers()

            cmd_run = subparser.add_parser("run", help="Start serving project")
            cmd_run.add_argument("-p", "--port", help="serving port", type=int, default=8080)
            cmd_run.set_defaults(handler=commands.run)
            cmd_auth= subparser.add_parser("auth", help="Authenticate user identified by screen name and password")
            cmd_auth.add_argument("-u", "--user", help="user's screen name", type=str)
            cmd_auth.add_argument("-p", "--password", help="password", type=str, nargs="?", default="", const="")
            cmd_auth.set_defaults(handler=commands.auth)
            cmd_uninstall = subparser.add_parser("uninstall", help="Uninstall app from project clearly")
            cmd_uninstall.add_argument("app", help="app's screen name")
            cmd_uninstall.set_defaults(handler=commands.uninstall)
        return self._arg_parser

    def command(self):
        args = self.arg_parser.parse_args()
        if hasattr(args, 'handler'):
            try:
                DatabaseManager.start_session()
                args.handler(self, args)
            except Exception:
                DatabaseManager.rollback_session()
            finally:
                DatabaseManager.close_session()
        else:
            self.arg_parser.print_help()

    def uninstall(self, screen_name):
        self.apps[screen_name].uninstall()

    def app(self, appname):
        return self.apps[appname]

    def wsgi(self, env, start_response):
        try:
            DatabaseManager.start_session()
            body = super().wsgi(env, start_response)
        except Exception as err:
            print_exc(err)
            DatabaseManager.rollback_session()
            request = Request(env)
            body = self.error(request, 500).start(request, start_response)
        finally:
            DatabaseManager.close_session()
        return body

def include(package, screen_name=None, project_dir=None, project_root_dir=None, path=None):
    if str(project_dir) not in sys.path:
        sys.path.append(str(project_dir))
    if package not in sys.modules:
        init = importlib.__import__(package, fromlist=["AppBuilder"])
    else:
        init = importlib.reload(package)
    builder = init.AppBuilder()
    if screen_name is None:
        screen_name = package
    builder.set_package(package)
    if screen_name is not None:
        builder.set_screen_name(screen_name)
    if project_root_dir is not None:
        builder.set_project_root_dir(project_root_dir)
    if project_dir is not None:
        builder.set_project_dir(project_dir)
    if path is not None:
        builder.set_path(path)
    return builder





#!/usr/bin/python

import os

from mitama.project import Project, include
from mitama.db import DatabaseManager

project_dir = os.path.dirname(os.path.abspath(__file__))

DatabaseManager({
    "type":"sqlite",
    "path": project_dir+'/db.sqlite3',
    #"host":"localhost",
    #"name":"mitama",
    #"user":"mitama",
    #"password":"mitama",
})

project = Project(
    include("mitama.portal", path="/"),
    project_dir = project_dir
)
application = project.wsgi


if __name__ == "__main__":
    project.command()



import copy
import inspect
import re


class RoutingError(Exception):
    pass


class Router:
    """ルーティングエンジン

    手軽に実装できて必要最低限なものを目指したので、遅かったりして嫌いだったら無理にこれを使う必要はありません。
    自前のものを適用したい場合は、とりあえず:samp:`hoge.match(Request): -> Response` といったインターフェースを実装したメソッドを作ってください。
    routesの中にRouterインスタンスを指定することもできます。
    """

    _app = None

    def __init__(self, routes=[], middlewares=[]):
        """初期化処理

        :param routes: Router、またはRouteのリスト
        :param middlewares: Middlewareのリスト
        """
        self._parent = None
        self.routes = list()
        self.middlewares = list()
        self.add_routes(routes)
        self.add_middlewares(middlewares)
        self.i = 0

    def add_route(self, route):
        """ルーティング先を追加します

        mitama.app.methodの関数で生成したRouteインスタンスを与えてください。
        :param route: Routeインスタンス
        """
        if isinstance(route, Router):
            route._parent = self
        self.routes.append(route)

    def add_routes(self, routes):
        """複数のルーティング先を追加します

        mitama.app.methodの関数で生成したRouteインスタンスを与えてください。
        :param routes: Routeインスタンスのリスト
        """
        for route in routes:
            self.add_route(route)

    def add_middleware(self, middleware):
        """ミドルウェアを登録します

        Middlewareクラスを与えると、このルーターのインスタンス内でマッチした場合にミドルウェアが順番に起動します。
        :param middleware: Middlewareのインスタンス
        """
        self.middlewares.append(middleware)

    def add_middlewares(self, middlewares):
        """複数のミドルウェアを登録します

        Middlewareクラスのリストを与えると、このルーターのインスタンス内でマッチした場合にミドルウェアが順番に起動します。
        :param middlewares: Middlewareのインスタンスのリスト
        """
        for middleware in middlewares:
            self.add_middleware(middleware)

    def clone(self):
        return Router(routes=copy.copy(self.routes))

    def match(self, request):
        method = request.method
        path = request.subpath if hasattr(request, "subpath") else request.path
        paths_to_check = [path]
        paths_to_check.append(re.sub("//+", "/", path))
        if not request.path.endswith("/"):
            paths_to_check.append(path + "/")
        paths_to_check.append(re.sub("//+", "/", path + "/"))
        if path.endswith("/"):
            merged_slashes = re.sub("//+", "/", path)
            paths_to_check.append(merged_slashes[:-1])

        for path in paths_to_check:
            for route in self.routes:
                request.subpath = path
                result = route.match(request)
                if result is not False:
                    request, result, method = result

                    def get_response_handler(result, method):
                        i = 0

                        def handle(request):
                            nonlocal i
                            nonlocal result
                            if inspect.isclass(result):
                                result = result(request.app)
                                if method is not None:
                                    inst = result

                                    def result(request):
                                        return inst(request, method)

                            if (
                                i >= len(self.middlewares) or
                                len(self.middlewares) == 0
                            ):
                                if callable(result):
                                    return result(request)
                                else:
                                    raise RoutingError(
                                        "Unsupported interface object. Only callables and Controller instances are supported."
                                    )
                            else:
                                if hasattr(request, "app"):
                                    middleware = self.middlewares[i](
                                        request.app
                                    )
                                else:
                                    middleware = self.middlewares[i]()
                                i += 1
                                return middleware.process(request, handle)

                        return handle

                    handler = get_response_handler(result, method)
                    return request, handler, None
        return False



import base64
import re
import ssl

from gevent.pywsgi import WSGIServer
from geventwebsocket.handler import WebSocketHandler

from .request import Request
from .response import Response


def run_app(app, port):
    server = WSGIServer(
        ("localhost", int(port)),
        app.wsgi,
        handler_class=WebSocketHandler
    )
    server.serve_forever()



import cgi
import http
import http.cookies
import io
import json
import importlib
import wsgiref.util as wsgiutil
from urllib.parse import parse_qs


class _Cookies:
    def __init__(self, cookies):
        self._cookies = cookies

    def __getitem__(self, key):
        if key in self._cookies:
            return self._cookies[key].value
        else:
            raise KeyError("")

    def get(self, key):
        if key in self._cookies:
            return self._cookies[key].value
        else:
            return None


class uWSGIWebSocket:
    def __init__(self, env):
        self.uwsgi = importlib.import_module("uwsgi")
        self.uwsgi.websocket_handshake(
            env['HTTP_SEC_WEBSOCKET_KEY'],
            env.get('HTTP_ORIGIN', '')
        )

    def receive(self):
        return self.uwsgi.websocket_recv()

    def send(self, message):
        self.uwsgi.websocket_send(message)


class _RequestPayload:
    def __init__(self, field_storage):
        self._field_storage = field_storage

    def __contains__(self, key):
        if key not in self._field_storage:
            return False
        elif isinstance(self._field_storage[key], list):
            return True
        elif isinstance(self._field_storage[key].file, io.BytesIO):
            return (self._field_storage[key].filename or "") != ""
        else:
            return len(self._field_storage[key].value) > 0

    def __getitem__(self, key):
        if key not in self._field_storage:
            raise KeyError
        elif (self._field_storage[key].filename or "") != "":
            return self._field_storage[key]
        elif len(self._field_storage[key].value) > 0:
            return self._field_storage[key].value
        else:
            raise KeyError

    def keys(self):
        return self._field_storage.keys()

    def values(self):
        return self._field_storage.values()

    def items(self):
        return self._field_storage.items()

    def get(self, key, default=None):
        if key in self:
            return self[key]
        else:
            return default

    def getlist(self, key, default=list()):
        if key in self:
            return self._field_storage.getlist(key)
        else:
            return default

    @classmethod
    def parse_body(cls, rfile, content_type, length):
        environ = {"REQUEST_METHOD": "POST"}
        headers = {"content-type": content_type, "content-length": length}
        fs = cgi.FieldStorage(rfile, environ=environ, headers=headers)
        payload = cls(fs)
        return payload


class Request:
    MessageClass = http.client.HTTPMessage

    def __init__(self, environ):
        self.environ = environ
        self._rfile = environ["wsgi.input"]
        self._data = dict()

    def __setitem__(self, key, value):
        self._data[key] = value

    def __getitem__(self, key):
        return self._data[key]

    def __delitem__(self, key):
        del self._data[key]

    def get(self, key):
        return self._data[key] if key in self._data else None

    @property
    def scheme(self):
        return wsgiutil.guess_scheme(self.environ)

    @property
    def host(self):
        return self.environ["HTTP_HOST"]

    @property
    def method(self):
        return self.environ["REQUEST_METHOD"].upper()

    @property
    def raw_path(self):
        return self.path + (
            "?" + "&".join(["%s=%s" % (kv[0], kv[1][0]) for kv in self.query.items()])
            if "QUERY_STRING" in self.environ
            else ""
        )

    @property
    def path(self):
        return self.environ.get("PATH_INFO", "/")

    @property
    def query(self):
        return parse_qs(self.environ.get("QUERY_STRING", ""))

    @property
    def version(self):
        return self.environ["SERVER_PROTOCOL"]

    @property
    def headers(self):
        return self.environ

    @property
    def url(self):
        return wsgiutil.request_uri(self.environ)

    @property
    def cookies(self):
        if hasattr(self, "_cookies"):
            return self._cookies
        else:
            cookie_header = self.environ.get("HTTP_COOKIE", "")
            C = http.cookies.SimpleCookie(cookie_header)
            self._cookies = _Cookies(C)
            return self._cookies

    @property
    def body(self):
        if hasattr(self, "_body"):
            return self._body
        else:
            length = self.environ.get("CONTENT_LENGTH", 0)
            self._body = self._rfile.read(int(length))
        return self._body

    @property
    def websocket(self):
        if hasattr(self, "_websocket"):
            return self._websocket
        else:
            if "wsgi.websocket" in self.environ and self.environ.get("wsgi.websocket") is not None:
                self._websocket = self.environ["wsgi.websocket"]
            else:
                try:
                    self._websocket = uWSGIWebSocket(self.environ)
                except Exception as err:
                    print(err)
        return self._websocket

    def post(self):
        if hasattr(self, "_post"):
            return self._post

        else:
            content_type = self.environ.get("CONTENT_TYPE", "")
            length = self.environ.get("CONTENT_LENGTH", 0)
            if length == "" or length is None:
                length = 0
            else:
                length = int(length)
            if content_type.startswith(
                "multipart/form-data"
            ) or content_type.startswith("application/x-www-form-urlencoded"):
                stream = io.BytesIO(self.body)
                parsed_body = _RequestPayload.parse_body(
                    stream, content_type, length
                )
            elif content_type == "application/json":
                payload = self.body
                parsed_body = json.loads(payload)
            else:
                parsed_body = {}
            self._post = parsed_body if parsed_body is not None else {}
            return self._post

    def session(self):
        """セッション情報を取得します

        :return: セッションの辞書データ
        """
        sess = self.get("mitama_session")
        if sess is None:
            storage = self.get("mitama_session_storage")
            sess = storage.load_session(self)
            self["mitama_session"] = sess
        return sess

    @classmethod
    def parse_stream(cls, rfile, ssl=False):
        words = rfile.readline().decode().rstrip("\r\n").split(" ")
        version = words[-1]
        if not 2 <= len(words) <= 3:
            raise Exception("Bad Request")
        command, path = words[:2]
        if len(words) == 2:
            if command != "GET":
                raise Exception("Bad Request")
        try:
            headers = http.client.parse_headers(rfile, _class=cls.MessageClass)
        except http.client.LineTooLong:
            raise Exception("")
        except http.client.HTTPException:
            raise Exception("")
        return cls(command, path, version, headers, ssl, rfile)



from mitama.app import Middleware
from mitama.app.http import Response
from mitama.models import User


class InitializeMiddleware(Middleware):
    def process(self, request, handler):
        if User.query.filter(User.password != None).count() == 0:
            return Response.redirect(self.app.convert_url("/setup"))
        else:
            return handler(request)



import os
import sys
import importlib
from mitama.app import AppRegistry
from mitama.app.http import run_app
from getpass import getpass


def run(project, args):
    port = args.port
    project.port = port
    run_app(project, project.port)

def auth(project, args):
    from mitama.models import User
    user = args.user
    password = args.password
    if password == "":
        password = getpass("Password:")
    try:
        user = User.password_auth(user, password)
        print(user.get_jwt())
        sys.exit(0)
    except Exception:
        print("Authentication failed")
        sys.exit(1)

def uninstall(project, args):
    project.uninstall(args.app)



#!/usr/bin/python
"""ベースモデル

    * アプリサイドから触るときに継承させるベースクラス
    * Djangoのdjango.db.modelsみたいなイメージ
    * データベースはユーザー管理をしているものとは別にしたいので、そのつもりで
    * UserとGroupみたいなカスタム型を使えるようになってると良い
    * Flaskのsqlalchemy拡張が参考に成る
"""

import uuid

from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import ColumnProperty, class_mapper
from sqlalchemy.types import TypeDecorator

from mitama._extra import _classproperty, tosnake

from .types import Column, Integer, String
from .event import Event


def UUID(prefix=None):
    def genUUID():
        s = str(uuid.uuid4())
        if prefix is not None:
            s = prefix + "-" + s
        return s
    return genUUID


class Model:
    prefix = None
    _id = Column(String(64), default=UUID(), primary_key=True, nullable=False)
    create = Event()
    update = Event()
    delete = Event()

    @classmethod
    def attribute_names(cls):
        return [
            prop.key
            for prop in class_mapper(cls).iterate_properties
            if isinstance(prop, ColumnProperty)
        ]

    def to_dict(self):
        attrs = self.attribute_names()
        d = dict()
        for k, v in self.__dict__.items():
            if k in attrs and (k == "_id" or k[0] != "_"):
                d[k] = v
        return d

    @_classproperty
    def type(cls):
        class Type(TypeDecorator):
            impl = Integer

            def process_bind_param(self, value, dialect):
                if value is None:
                    return None
                else:
                    return value._id

            def process_result_value(self, value, dialect):
                if value is None:
                    return None
                else:
                    user = cls.retrieve(value)
                    return user

        return Type

    @declared_attr
    def __tablename__(cls):
        return (
            ""
            if cls.prefix is None
            else cls.prefix + "_"
        ) + tosnake(cls.__name__)

    def create(self):
        self.query.session.add(self)
        self.query.session.commit()
        try:
            self.on("create")()
        except Exception:
            pass

    def update(self):
        self.query.session.commit()
        try:
            self.on("update")()
        except Exception:
            pass

    def delete(self):
        try:
            self.on("delete")()
        except Exception:
            pass
        self.query.session.delete(self)
        self.query.session.commit()

    def on(self, evt):
        return getattr(self, evt)

    @classmethod
    def listen(cls, evt):
        setattr(cls, evt, Event())

    @classmethod
    def list(cls, *args):
        if len(args) > 0:
            return cls.query.filter(args).all()
        else:
            return cls.query.filter().all()

    @classmethod
    def retrieve(cls, id=None, **kwargs):
        if id is not None:
            node = cls.query.filter(cls._id == id).one()
        elif len(kwargs) > 0:
            q = cls.query
            for attr, value in kwargs.items():
                q = q.filter(getattr(cls, attr) == value)
            node = q.one()
        else:
            raise Exception("Identity not given")
        return node



from mitama.app.forms import Form, FileField, Field, DictField, empty_error, format_error
from mitama.noimage import load_noimage_group, load_noimage_user

class SetupForm(Form):
    email = Field(label="メールアドレス", required=True)

class SettingsForm(Form):
    welcome_message = Field(label="ウェルカムボード", required=True)
    role_name = Field(label='役割名')
    role_screen_name = Field(label='役割ID', regex=r'^([0-9a-zA-Z_-]+)$')
    permission = DictField(listed=True)
    inner_role_name = Field(label='グループ役割名')
    inner_role_screen_name = Field(label='グループ役割ID', regex=r'^([0-9a-zA-Z_-]+)$')
    inner_permission = DictField(listed=True)

class LoginForm(Form):
    screen_name = Field(label="ログイン名", required=True)
    password = Field(label="パスワード", required=True)


class RegisterForm(Form):
    icon = FileField(label="プロフィール画像", initial=load_noimage_user())
    screen_name = Field(label="ログイン名", required=True)
    name = Field(label="名前", required=True)
    password = Field(label="パスワード", required=True)


class InviteForm(Form):
    email = Field(label="メールアドレス", required=True)
    icon = FileField(label="プロフィール画像", initial=load_noimage_user())
    name = Field(label="名前")
    roles = Field(label="役割", listed=True)
    screen_name = Field(label="ログイン名")
    editable = Field(label="プロフィールの変更を許可", initial=False)


class UserUpdateForm(Form):
    icon = FileField(label="プロフィール画像")
    name = Field(label="名前", required=True)
    screen_name = Field(label="ログイン名")
    roles = Field(label="役割", listed=True)

class SubscriptionForm(Form):
    action = Field(required=True)
    subscription = Field()

class UserPasswordUpdateForm(Form):
    password = Field(label="パスワード", required=True)
    password_ = Field(label="確認用パスワード", required=True)

class GroupCreateForm(Form):
    icon = FileField(label="プロフィール画像", initial=load_noimage_group())
    name = Field(label="名前", required=True)
    screen_name = Field(label="ドメイン名", required=True)
    parent = Field(label="親グループ")


class GroupUpdateForm(Form):
    icon = FileField(label="プロフィール画像")
    name = Field(label="名前", required=True)
    screen_name = Field(label="ドメイン名", required=True)
    parent = Field(label="親グループ")
    roles = Field(label="役割", listed=True)
    users = Field(listed=True)
    inner_roles = DictField(listed=True)
    new_user = Field()


class AppUpdateForm(Form):
    prefix = DictField(label="配信先", required=True)



#!/usr/bin/python
"""コマンド定義

    * 1ファイル1コマンドでクラスとして書いてます
    * handleがコマンドライン引数の引数部を受け取って実行する
    * 抽象化はするかどうか迷ってる（Pythonで抽象化がどれくらい尊いものかよくわかんない）
    * argparseを入れようか迷ったけど、位置引数しか必要に成る予定が無いのでとりあえず入れてない
"""
import warnings
warnings.simplefilter("ignore")

import argparse
import mitama.commands
import gevent.monkey
#import sqlalchemy_gevent

gevent.monkey.patch_all()

parser = argparse.ArgumentParser(description="Mitama command utilities")
subparser = parser.add_subparsers()

init = subparser.add_parser("init", help="Initialize directory as a Mitama project")
init.set_defaults(handler=commands.init)

new = subparser.add_parser("new", help="Create new Mitama project directory")
new.add_argument("name", help="new project's name")
new.set_defaults(handler=commands.new)

mkapp = subparser.add_parser("mkapp", help="Create new Mitam application")
mkapp.add_argument("name", help="new app's name")
mkapp.set_defaults(handler=commands.mkapp)

version = subparser.add_parser("version", help="Check version code")
version.set_defaults(handler=commands.version)

def command_exec():
    """コマンドを起動します

    コマンドライン引数からサブコマンドのインスタンスを生成し、起動します。
    """
    args = parser.parse_args()
    args.handler(args)



import urllib
import base64
from secrets import token_hex

from mitama.app import Middleware
from mitama.app.http import Response
from mitama.models import User


class SessionMiddleware(Middleware):
    """ログイン判定ミドルウェア

    ログインしていないユーザーがアクセスした場合、/login?redirect_to=<URL>にリダイレクトします。
    """

    def process(self, request, handler):
        sess = request.session()
        try:
            if "jwt_token" in sess:
                request.user = User.check_jwt(sess["jwt_token"])
            else:
                return Response.redirect(
                    "/login?redirect_to="
                    + urllib.parse.quote(str(request.url), safe="")
                )
        except Exception as err:
            return Response.redirect(
                "/login?redirect_to=" + urllib.parse.quote(str(request.url), safe="")
            )
        return handler(request)

class BasicMiddleware(Middleware):
    """BASIC認証ミドルウェア"""

    def process(self, request, handler):
        try:
            if "HTTP_AUTHORIZATION" in request.headers:
                name, token = request.headers["HTTP_AUTHORIZATION"].split(" ")
                login, password = base64.b64decode(token).decode().split(":")
                request.user = User.password_auth(login, password)
            else:
                return Response(status=401, reason="Authorization Required", headers = {
                    "WWW-Authenticate": "Basic realm=\"mitama authorization\""
                })
        except Exception as err:
            print(err)
            return Response(status=401, reason="Authorization Required", headers = {
                "WWW-Authenticate": "Basic realm=\"mitama authorization\""
            })
        return handler(request)

class CsrfMiddleware(Middleware):
    def process(self, request, handler):
        sess = request.session()
        if request.method == "POST":
            session_token = sess["mitama_csrf_token"]
            posts = request.post()
            if "mitama_csrf_token" not in posts:
                return Response(status=400, reason="Bad Request")
            post_token = posts["mitama_csrf_token"]
            if session_token != post_token:
                return Response(status=400, reason="Bad Request")
        else:
            sess["mitama_csrf_token"] = token_hex(16)
        return handler(request)



import os
from mimetypes import add_type, guess_type
from pathlib import Path

import magic
from jinja2 import *

from mitama.app import Controller
from mitama.app.http import Request, Response
from mitama.models import Group, User

add_type("application/json", ".map")


def static_files(*paths):
    paths_ = list(paths)

    class StaticFileController(Controller):
        """静的ファイルを配信するController

        デフォルトではアプリのパッケージ内の :file:`static/` の中身を配信する。
        """

        paths = paths_

        def __init__(self, app):
            super().__init__(app)
            app_mod_dir = Path(os.path.dirname(__file__))
            self.view = Environment(
                loader=FileSystemLoader(
                    [app_mod_dir / "templates", app_mod_dir / "../app/templates"]
                )
            )
            if len(self.paths) == 0:
                self.paths.append(self.app.install_dir / "static")

        def handle(self, req: Request):
            for path in self.paths:
                filename = path / req.params["path"]
                if filename.is_file():
                    mime = guess_type(str(filename)) or ("application/octet-stream",)
                    with open(filename, "rb") as f:
                        return Response(body=f.read(), content_type=mime[0])
            for path in self.paths:
                filename = path / "404.html"
                if filename.is_file():
                    with open(filename) as f:
                        return Response(
                            text=f.read(),
                            status=404,
                            headers={"content-type": "text/html"},
                        )
            template = self.view.get_template("404.html")
            return Response.render(template, status=404)

    return StaticFileController


def mitama_favicon(*paths):
    paths_ = list(paths)

    class FaviconController(Controller):
        """静的ファイルを配信するController

        デフォルトではアプリのパッケージ内の :file:`static/` の中身を配信する。
        """

        paths = paths_

        def __init__(self, app):
            super().__init__(app)
            app_mod_dir = Path(os.path.dirname(__file__))
            self.view = Environment(
                loader=FileSystemLoader(
                    [app_mod_dir / "templates", app_mod_dir / "../app/templates"]
                )
            )
            if len(self.paths) == 0:
                self.paths.append(app_mod_dir / "../app/static")

        def handle(self, req: Request):
            for path in self.paths:
                filename = path / "favicon.ico"
                if filename.is_file():
                    mime = guess_type(str(filename)) or ("application/octet-stream",)
                    with open(filename, "rb") as f:
                        return Response(body=f.read(), content_type=mime[0])
            for path in self.paths:
                filename = path / "404.html"
                if filename.is_file():
                    with open(filename) as f:
                        return Response(
                            text=f.read(),
                            status=404,
                            headers={"content-type": "text/html"},
                        )
            template = self.view.get_template("404.html")
            return Response.render(template, status=404)

    return FaviconController

def mitama_service_worker(*paths):
    paths_ = list(paths)

    class ServiceWorkerController(Controller):
        """静的ファイルを配信するController

        デフォルトではアプリのパッケージ内の :file:`static/` の中身を配信する。
        """

        paths = paths_

        def __init__(self, app):
            super().__init__(app)
            app_mod_dir = Path(os.path.dirname(__file__))
            self.view = Environment(
                loader=FileSystemLoader(
                    [app_mod_dir / "templates", app_mod_dir / "../app/templates"]
                )
            )
            if len(self.paths) == 0:
                self.paths.append(app_mod_dir / "../app/static")

        def handle(self, req: Request):
            for path in self.paths:
                filename = path / "sw.js"
                if filename.is_file():
                    mime = guess_type(str(filename)) or ("application/octet-stream",)
                    with open(filename, "rb") as f:
                        return Response(body=f.read(), content_type=mime[0])
            for path in self.paths:
                filename = path / "404.html"
                if filename.is_file():
                    with open(filename) as f:
                        return Response(
                            text=f.read(),
                            status=404,
                            headers={"content-type": "text/html"},
                        )
            template = self.view.get_template("404.html")
            return Response.render(template, status=404)

    return ServiceWorkerController

def mitama_manifest():
    class ManifestController(Controller):
        def handle(self, req: Request):
            manifest = {
                "name": "Mitama",
                "short_name": "Mitama",
                "icons": [{
                    "src": self.app.convert_fullurl(req, "/static/icon.png"),
                    "sizes": "243x243",
                    "type": "image/png"
                }],
                "start_url": "/",
                "display": "standalone"
            }
            return Response.json(manifest)

    return ManifestController


class UserCRUDController(Controller):
    def create(self, request):
        post = request.post()
        try:
            user = User()
            user.screen_name = post["screen_name"]
            user.name = post["name"]
            user.set_password(post["password"])
            user.create()
            return Response.json(user.to_dict())
        except KeyError as err:
            error = err
            return Response.json(
                {
                    "error": err.__class__.__name__,
                    "message": err.message,
                    "args": err.args,
                }
            )

    def retrieve(self, request):
        try:
            id = request.params["id"]
            if id.isdigit():
                user = User.retrieve(int(id))
            else:
                user = User.retrieve(screen_name=id)
            return Response.json(user.to_dict())
        except KeyError as err:
            error = err
            return Response.json(
                {
                    "error": err.__class__.__name__,
                    "message": err.message,
                    "args": err.args,
                }
            )

    def icon(self, request):
        try:
            id = request.params["id"]
            if id.isdigit():
                user = User.retrieve(int(id))
            else:
                user = User.retrieve(screen_name=id)
            f = magic.Magic(mime=True, uncompress=True)
            mime = f.from_buffer(group.icon)
            return Response(body=user.icon, content_type=mime)
        except KeyError as err:
            return Response.error(
                {
                    "error": err.__class__.__name__,
                    "message": err.message,
                    "args": err.args,
                }
            )

    def update(self, request):
        try:
            post = request.post()
            id = request.params["id"]
            if id.isdigit():
                user = User.retrieve(int(id))
            else:
                user = User.retrieve(screen_name=id)
            if "screen_name" in post:
                user.screen_name = post["screen_name"]
            if "name" in post:
                user.name = post["name"]
            if "password" in post:
                user.set_password(post["password"])
            user.update()
            return Response.json(user.to_dict())
        except KeyError as err:
            error = err
            return Response.json(
                {
                    "error": err.__class__.__name__,
                    "message": err.message,
                    "args": err.args,
                }
            )

    def delete(self, request):
        try:
            id = request.params["id"]
            if id.isdigit():
                user = User.retrieve(int(id))
            else:
                user = User.retrieve(screen_name=id)
            user.delete()
            return Response.json({"_id": user._id})
        except Exception as err:
            error = err
            return Response.json(
                {
                    "error": err.__class__.__name__,
                    "message": err.message,
                    "args": err.args,
                }
            )

    def list(self, request):
        users = User.list()
        return Response.json([user.to_dict for user in users])


class GroupCRUDController(Controller):
    def create(self, request):
        post = request.post()
        try:
            group = Group()
            group.screen_name = post["screen_name"]
            group.name = post["name"]
            group.create()
            return Response.json(group.to_dict())
        except KeyError as err:
            error = err
            return Response.json(
                {
                    "error": err.__class__.__name__,
                    "message": err.message,
                    "args": err.args,
                }
            )

    def retrieve(self, request):
        try:
            id = request.params["id"]
            if id.isdigit():
                group = Group.retrieve(int(id))
            else:
                group = Group.retrieve(screen_name=id)
            return Response.json(group.to_dict())
        except KeyError as err:
            error = err
            return Response.json(
                {
                    "error": err.__class__.__name__,
                    "message": err.message,
                    "args": err.args,
                }
            )

    def icon(self, request):
        try:
            id = request.params["id"]
            if id.isdigit():
                group = Group.retrieve(int(id))
            else:
                group = Group.retrieve(screen_name=id)
            f = magic.Magic(mime=True, uncompress=True)
            mime = f.from_buffer(group.icon)
            return Response(body=group.icon, content_type=mime)
        except KeyError as err:
            return Response.error(
                {
                    "error": err.__class__.__name__,
                    "message": err.message,
                    "args": err.args,
                }
            )

    def update(self, request):
        try:
            post = request.post()
            id = request.params["id"]
            if id.isdigit():
                group = Group.retrieve(int(id))
            else:
                group = Group.retrieve(screen_name=id)
            if "screen_name" in post:
                group.screen_name = post["screen_name"]
            if "name" in post:
                group.name = post["name"]
            group.update()
            return Response.json(group.to_dict())
        except KeyError as err:
            error = err
            return Response.json(
                {
                    "error": err.__class__.__name__,
                    "message": err.message,
                    "args": err.args,
                }
            )

    def delete(self, request):
        try:
            id = request.params["id"]
            if id.isdigit():
                group = Group.retrieve(int(id))
            else:
                group = Group.retrieve(screen_name=id)
            group.delete()
            return Response.json({"_id": group._id})
        except Exception as err:
            error = err
            return Response.json(
                {
                    "error": err.__class__.__name__,
                    "message": err.message,
                    "args": err.args,
                }
            )

    def list(self, request):
        groups = Group.list()
        return Response.json([group.to_dict() for group in groups])



#!/usr/bin/python
"""コマンド定義

    * 1ファイル1コマンドでクラスとして書いてます
    * handleがコマンドライン引数の引数部を受け取って実行する
    * 抽象化はするかどうか迷ってる（Pythonで抽象化がどれくらい尊いものかよくわかんない）
    * argparseを入れようか迷ったけど、位置引数しか必要に成る予定が無いのでとりあえず入れてない
"""

import importlib
import sys
import argparse
import os
import shutil
import glob
import pkg_resources
from pathlib import Path


def init_project_dir(path):
    files = glob.glob(os.path.dirname(__file__) + "/skeleton/project/*")
    for file in files:
        if os.path.isdir(file):
            continue
        shutil.copy(file, path, follow_symlinks=False)


def init(args):
    path = Path(os.getcwd())
    init_project_dir(path)

def mkapp(args):
    try:
        project_name = args.name
    except IndexError:
        raise IndexError("No app name given to command arguments.")
    current_dir = Path(os.getcwd())
    project_dir = current_dir / project_name
    src = Path(os.path.dirname(__file__)) / "skeleton/app_templates"
    shutil.copytree(src, project_dir, symlinks=False)

def new(args):
    try:
        project_name = args.name
    except IndexError:
        raise IndexError("No project name given to command arguments.")
    current_dir = Path(os.getcwd())
    project_dir = current_dir / project_name
    os.mkdir(project_dir)
    init_project_dir(project_dir)

def version(args):
    v = pkg_resources.get_distribution("mitama").version
    print(v)




#!/usr/bin/python
"""sqliteのエンジン

    * プロジェクトのDBはプロジェクトのconfigから場所を取って参照する。
    * アプリのDBはアプリのフォルダ直下を参照する
"""

from sqlalchemy import create_engine


def get_engine():
    from mitama.conf import get_from_project_dir

    config = get_from_project_dir()
    return create_engine("sqlite:///" + str(config._sqlite_db_path))


def get_test_engine():
    return create_engine("sqlite://")




from mitama.app import Router
from mitama.app.method import *

from .controllers import GroupCRUDController, UserCRUDController
from .middlewares import SessionMiddleware


def create_api_router(middlewares=[SessionMiddleware]):
    return Router(
        [
            post("/users", UserCRUDController, "create"),
            post("/groups", GroupCRUDController, "create"),
        ],
        middlewares,
    )


def retrieve_api_router(middlewares=[]):
    return Router(
        [
            get("/users", UserCRUDController, "list"),
            get("/users/<id>", UserCRUDController, "retrieve"),
            get("/users/<id>/icon", UserCRUDController, "icon"),
            get("/groups", GroupCRUDController, "list"),
            get("/groups/<id>", GroupCRUDController, "retrieve"),
            get("/groups/<id>/icon", GroupCRUDController, "icon"),
        ],
        middlewares,
    )


def update_api_router(middlewares=[SessionMiddleware]):
    return Router(
        [
            post("/users/<id>", UserCRUDController, "update"),
            post("/groups/<id>", GroupCRUDController, "update"),
        ],
        middlewares,
    )


def delete_api_router(middlewares=[SessionMiddleware]):
    return Router(
        [
            delete("/users/<id>", UserCRUDController, "delete"),
            delete("/groups/<id>", GroupCRUDController, "delete"),
        ],
        middlewares,
    )


def rest_api_router(middlewares=[SessionMiddleware]):
    return Router(
        [
            post("/users", UserCRUDController, "create"),
            post("/groups", GroupCRUDController, "create"),
            get("/users", UserCRUDController, "list"),
            get("/users/<id>", UserCRUDController, "retrieve"),
            get("/groups", GroupCRUDController, "list"),
            get("/groups/<id>", GroupCRUDController, "retrieve"),
            post("/users/<id>", UserCRUDController, "update"),
            post("/groups/<id>", GroupCRUDController, "update"),
            delete("/users/<id>", UserCRUDController, "delete"),
            delete("/groups/<id>", GroupCRUDController, "delete"),
        ],
        middlewares,
    )



import os
from base64 import b64encode
from pathlib import Path

import magic


def load_noimage_app():
    """アプリのNoImage画像を取得します"""
    noimage_app_path = (
        Path(os.path.dirname(__file__)) / "app/static/noimage_app.png"
    )
    with open(noimage_app_path, "rb") as f:
        noimage_app = f.read()
    return noimage_app


def load_noimage_user():
    """ユーザーのNoImage画像を取得します"""
    noimage_user_path = (
        Path(os.path.dirname(__file__)) / "app/static/noimage_user.png"
    )

    with open(noimage_user_path, "rb") as f:
        noimage_user = f.read()
    return noimage_user


def load_noimage_group():
    """グループのNoImage画像を取得します"""
    noimage_group_path = (
        Path(os.path.dirname(__file__)) / "app/static/noimage_group.png"
    )

    with open(noimage_group_path, "rb") as f:
        noimage_group = f.read()
    return noimage_group



#!/usr/bin/python
"""データベース

データベースの接続とか抽象化の処理を書きます
Databaseはシングルトンの接続のインスタンスを生成するクラスです
各アプリにはDatabaseを継承したクラスを定義してもらい、そいつのModelプロパティのベースクラスからモデルを作ってもらいます。
"""

import inspect as _inspect

from sqlalchemy import *
from sqlalchemy.engine import *
from sqlalchemy.schema import *
from sqlalchemy.inspection import inspect
from sqlalchemy.sql import *
from sqlalchemy.types import *
from sqlalchemy.orm import *
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.declarative.api import DeclarativeMeta
from sqlalchemy.orm.exc import UnmappedClassError
from sqlalchemy.sql import func

from mitama._extra import _Singleton

from .driver.sqlite3 import get_test_engine
from .model import Model


class _QueryProperty:
    def __init__(self, db):
        self.db = db

    def __get__(self, obj, type):
        try:
            mapper = class_mapper(type)
            if mapper:
                return Query([type]).with_session(self.db.session)
        except UnmappedClassError:
            return None


class DatabaseManager(_Singleton):
    engine = None
    metadata = None
    session = None

    @classmethod
    def test(cls):
        cls.set_engine(get_test_engine())

    @classmethod
    def set_engine(cls, engine):
        cls.engine = engine
        cls.metadata = MetaData(cls.engine)
        cls.Model = declarative_base(
            cls=Model,
            name="Model",
            metadata=cls.metadata
        )

    @classmethod
    def start_session(cls):
        cls.session = Session(autocommit=False, autoflush=False, bind=cls.engine)

    @classmethod
    def close_session(cls):
        cls.session.close()

    @classmethod
    def rollback_session(cls):
        cls.session.rollback()

    def __init__(self, database=None):
        if database is not None:
            if database["type"] == "mysql":
                engine = create_engine(
                    "mysql://{}:{}@{}/{}?charset=utf8mb4".format(
                        database["user"],
                        database["password"],
                        database["host"],
                        database["name"]
                    )
                )
            elif database["type"] == "postgresql":
                engine = create_engine(
                    "postgresql://{}:{}@{}/{}?charset=utf8".format(
                        database["user"],
                        database["password"],
                        database["host"],
                        database["name"]
                    ),
                    echo=True
                )
            else:
                engine = create_engine("sqlite:///" + str(database["path"]))
            self.set_engine(engine)


class _Database():
    def __init__(self, model=None, metadata=None, query_class=Query):
        self.manager = DatabaseManager()
        self.Query = query_class
        self.Model = self.make_declarative_base(model, metadata)

    def make_declarative_base(self, model=None, metadata=None):
        if model is None:
            model = self.manager.Model
        if not isinstance(model, DeclarativeMeta):
            model = declarative_base(
                cls=model, name="Model", metadata=metadata
            )
        if metadata is not None and model.metadata is not metadata:
            model.metadata = metadata
        else:
            model.metadata = self.metadata
        if not getattr(model, "query_class", None):
            model.query_class = self.Query
        model.query = _QueryProperty(self)
        return model

    @property
    def engine(self):
        return self.manager.engine

    @property
    def metadata(self):
        return self.manager.metadata

    @property
    def session(self):
        return self.manager.session

    def create_all(self):
        self.metadata.create_all(self.engine)


class BaseDatabase(_Database):
    """アプリで利用するデータベースの操作を行うクラス

    アプリからデータベースを使うたい場合、このクラスを継承したクラスをアプリ内に定義します。
    """

    def __init__(self, prefix=None, model=None, metadata=None, query_class=Query):
        super().__init__(
            model=model,
            metadata=metadata,
            query_class=query_class
        )
        if prefix is None:
            prefix = _inspect.getmodule(self.__class__).__package__
        self.Model.prefix = prefix



from jinja2 import Template

class ValidationError(Exception):
    template = Template("{{ label }}が適切ではありません")
    def __init__(self, label="", data=""):
        self.label = label
        self.data = data

    @classmethod
    def setTemplate(cls, template = "{{ label }}が適切ではありません"):
        cls.template = Template(template)

    @property
    def message(self):
        return self.template.render(
            label = self.label,
            data = self.data
        )

    def __str__(self):
        return self.message

class EmptyError(ValidationError):
    template = Template("{{ label }}が入力されていません")
    def __init__(self, label=""):
        super().__init__(label, None)

class FormatError(ValidationError):
    template = Template("{{ label }}の形式が間違っています")
    def __init__(self, label="", data=""):
        super().__init__(label, data)

def empty_error(template_string):
    class EmtpyError_(EmptyError):
        template = Template(template_string)
    return EmtpyError_

def format_error(template_string):
    class FormatError_(FormatError):
        template = Template(template_string)
    return FormatError_



import http.cookies
import json
import types
from abc import ABCMeta, abstractmethod
from http.server import BaseHTTPRequestHandler


class ResponseBase(metaclass=ABCMeta):
    _version = "HTTP/1.1"

    def __init__(self, status=200, reason=None, headers={}):
        self.headers = headers
        self._status = int(status)
        if reason is None:
            try:
                self._reason = BaseHTTPRequestHandler.responses[self._status][0]
            except KeyError:
                self._reason = ""
        else:
            self._reason = reason
        self._cookies = http.cookies.SimpleCookie()

    def set_cookie(self, key, value=None, **kwargs):
        self._cookies[key] = value or ""
        for k, v in kwargs.items():
            self._cookies[key][k] = v or ""

    @abstractmethod
    def start(self, request, stream):
        pass


class Response(ResponseBase):
    def __init__(
        self,
        body=None,
        text=None,
        headers={},
        status=200,
        reason=None,
        content_type=None,
        charset=None,
    ):
        super().__init__(headers=headers, status=status, reason=reason)
        if body is not None and text is not None:
            raise ValueError("body and text are not allowed together")
        if text is not None:
            if not isinstance(text, str):
                raise TypeError("")
            if content_type is None:
                content_type = "text/html"
            if charset is None:
                charset = "utf-8"
            body = text.encode(charset)
            text = None
        else:
            if content_type is not None:
                if charset is not None:
                    content_type += "; charset=" + charset
            else:
                content_type = "text/html"
        self.content_type = content_type
        self.body = body

    def start(self, request, start_response):
        headers = list()
        for kv in self.headers.items():
            headers.append(kv)
        cookies = self._cookies.output(header="")
        if len(cookies) > 0:
            headers.extend([("Set-Cookie", cookie) for cookie in cookies.split("\r\n")])
        headers.append(("Content-Type", self.content_type))
        start_response(("%s %s" % (self._status, self._reason)), headers)
        if callable(self.body):
            return [self.body(request)]
        elif isinstance(self.body, types.GeneratorType):
            return [*self.body]
        elif self.body is not None:
            return [self.body]
        else:
            return []

    @classmethod
    def json(cls, d, **kwargs):
        if "content_type" not in kwargs:
            kwargs["content_type"] = "application/json"
        return cls(body=json.dumps(d).encode(), **kwargs)

    @classmethod
    def render(cls, template, values={}, **kwargs):
        """HTMLを描画するレスポンスを返却します

        Jinja2のテンプレートにデータを入れてHTMLを生成し、Responseインスタンスを作成して返却します。
        テンプレート内部からはRequest内の非同期関数を呼び出すことができます。
        :param template: Jinja2テンプレート
        :param request: Requestインスタンス
        :param values: プレースホルダに入力するデータの辞書
        :param **kwargs: aiohttp.web.Responseに受け渡す、statusやheadersなどのデータ
        """

        def render(request):
            values["request"] = request
            return template.render(values).encode()

        return cls(body=render, **kwargs)

    @classmethod
    def redirect(cls, uri, status=302):
        """リダイレクトします

        :param uri: リダイレクト先のURL
        :param status: リダイレクト時のステータスコード
        """
        return cls(headers={"Location": uri}, status=status)


class StreamResponse(ResponseBase):
    def __init__(
        self,
        body=None,
        headers={},
        status=200,
        reason=None,
        content_type=None,
        charset=None,
    ):
        super().__init__(headers=headers, status=status, reason=reason)
        self.content_type = content_type
        if charset is not None:
            self.content_type += "; charset=" + charset
        if body is not None:
            self.body = body
        else:
            self.body = list()

    def start_wsgi(self, request, start_response):
        headers = list()
        for kv in self.headers.items():
            headers.append(kv)
        cookies = self._cookies.output(attrs=[], header="")
        if len(cookies) > 0:
            headers.extend([("Set-Cookie", cookie) for cookie in cookies.strip("\r\n")])
        headers.append(("Content-Type", self.content_type))

        def content():
            nonlocal self, start_response, request, headers
            start_response(("%s %s" % (self._status, self._reason)), headers)
            for chunk in self.body:
                if callable(chunk):
                    yield chunk(request)
                elif chunk is not None:
                    yield chunk
        return content()

    def start(self, request, stream):
        stream.write(
            ("%s %s %s\r\n" % (self._version, self._status, self._reason)).encode()
        )
        for k, v in self.headers.items():
            stream.write(("%s: %s\r\n" % (k, v)).encode())
        stream.write(self._cookies.output().encode())
        stream.write(("Content-Type: %s\r\n" % self.content_type).encode())
        stream.write("\r\n".encode())
        for chunk in self.body:
            if callable(chunk):
                stream.write(chunk())
            else:
                stream.write(chunk)



#!/usr/bin/python
"""型定義

    * ノリで書いたから自分でもよくわからないけど、データベースのカスタム型を定義してみた
    * イメージとしては、データベースの実体には各モデルのidプロパティが入る
    * 取り出すときは数字からそれに対応するモデルのインスタンスが還ってくる
    * 当然、↓のコードはUser.getなんて関数は用意してないので動かないと思われ

"""

from sqlalchemy import *
from sqlalchemy.types import TypeDecorator


class User(TypeDecorator):
    impl = Integer

    def process_bind_param(self, value, dialect):
        if value == None:
            return None
        elif value.__class__.__name__ == "User":
            return value._id
        else:
            return value

    def process_result_value(self, value, dialect):
        from mitama.models import User

        if value == None:
            return None
        else:
            user = User.retrieve(value)
            return user


class Group(TypeDecorator):
    impl = Integer

    def process_bind_param(self, value, dialect):
        if value == None:
            return None
        elif value.__class__.__name__ == "Group":
            return value._id
        else:
            return value

    def process_result_value(self, value, dialect):
        from mitama.models import Group

        if value == None:
            return None
        else:
            group = Group.retrieve(value)
            return group


class Node(TypeDecorator):
    impl = Integer

    def process_bind_param(self, value, dialect):
        if value.__class__.__name__ == "Group" or value.__class__.__name__ == "User":
            return value._id
        elif type(value) == str and value.split("-")[0] in ["user", "group"]:
            return value
        else:
            raise TypeError("Appending object must be Group or User instance")

    def process_result_value(self, value, dialect):
        prefix = value.split("-")[0]
        if prefix == "user":
            from mitama.models import User

            node = User.retrieve(value)
        elif prefix == "group":
            from mitama.models import Group

            node = Group.retrieve(value)
        else:
            raise TypeError("Invalid ID format")
        return node



import json
import traceback
import io
from uuid import uuid4

from mitama.app import AppRegistry, Controller
from mitama.app.http import Response
from mitama.models import (
    AuthorizationError,
    Group,
    User,
    Role,
    UserInvite,
    InnerRole,
    Permission,
    InnerPermission,
    UserGroup,
    PushSubscription
)
from mitama.app.forms import ValidationError
from mitama.noimage import load_noimage_group, load_noimage_user

from .forms import (
    SetupForm,
    LoginForm,
    RegisterForm,
    InviteForm,
    UserUpdateForm,
    UserPasswordUpdateForm,
    SubscriptionForm,
    GroupCreateForm,
    GroupUpdateForm,
    AppUpdateForm,
    SettingsForm
)

from PIL import Image

def resize_icon(icon):
    if icon is None:
        return None
    try:
        img = Image.open(io.BytesIO(icon))
    except Exception:
        return icon
    width, height = img.size
    if width > height:
        scale = 200 / height
    else:
        scale = 200 / width
    width *= scale
    height *= scale
    resize = img.resize((int(width), int(height)), resample=Image.NEAREST)
    if width > height:
        cropped = resize.crop((int((width-height)/2), 0, width - int((width-height)/2), height))
    else:
        cropped = resize.crop((0, int((height-width)/2), width, height - int((height-width)/2)))
    export = io.BytesIO()
    cropped.save(export, format="PNG")
    return export.getvalue()

class SessionController(Controller):
    def login(self, request):
        template = self.view.get_template("login.html")
        if request.method == "POST":
            try:
                form = LoginForm(request.post())
                result = User.password_auth(
                    form["screen_name"],
                    form["password"]
                )
                sess = request.session()
                sess["jwt_token"] = result.get_jwt()
                redirect_to = request.query.get("redirect_to", ["/"])[0]
                return Response.redirect(redirect_to)
            except Exception as err:
                return Response.render(template, {"error": err}, status=401)
        return Response.render(template, status=401)

    def logout(self, request):
        sess = request.session()
        sess["jwt_token"] = None
        redirect_to = request.query.get("redirect_to", ["/"])[0]
        return Response.redirect(redirect_to)


class RegisterController(Controller):
    def signup(self, request):
        sess = request.session()
        template = self.view.get_template("signup.html")
        invite = UserInvite.retrieve(token=request.query["token"][0])
        if request.method == "POST":
            try:
                form = RegisterForm(request.post())
                user = User()
                user.email = invite.email
                user.set_password(form["password"])
                user.screen_name = form["screen_name"]
                user.name = form["name"]
                user.icon = resize_icon(form["icon"]) if form["icon"] is not None else user.icon
                user.create()
                sess["jwt_token"] = user.get_jwt()
                roles = invite.roles;
                if len(roles) > 0:
                    for role_screen_name in invite.roles.split(":"):
                        role = Role.retrieve(screen_name = role_screen_name)
                        role.append(user)
                invite.delete()
                return Response.redirect(self.app.convert_url("/"))
            except (ValidationError, ValueError) as err:
                icon = form["icon"]
                return Response.render(
                    template,
                    {
                        "error": err.messsage,
                        "name": form["name"] or invite.name,
                        "screen_name": form["screen_name"] or invite.screen_name,
                        "password": form["password"] or "",
                        "icon": icon,
                    },
                )
        return Response.render(
            template,
            {
                "icon": invite.icon,
                "name": invite.name or "",
                "screen_name": invite.screen_name or "",
            },
        )

    def setup(self, request):
        sess = request.session()
        template = self.view.get_template("setup.html")
        if len(User.list()) > 0:
            return Response.redirect(self.app.convert_url("/login"))
        if request.method == "POST":
            try:
                form = SetupForm(request.post())
                user = UserInvite()
                user.email = form["email"]
                user.roles = "owner"
                user.create()
                user.mail(
                    "Mitamaへようこそ",
                    "下記リンクから、Mitamaに参加しましょう\n{}".format(self.app.convert_fullurl(request, "/signup?token=" + user.token))
                )

                try:
                    owner = Role()
                    owner.screen_name = "owner"
                    owner.name = "Owner"
                    owner.create()
                    manager = Role()
                    manager.screen_name = "manager"
                    manager.name = "Manager"
                    manager.create()
                    normal = Role()
                    normal.screen_name = "normal"
                    normal.name = "Normal"
                    normal.create()
                    inner_owner = InnerRole()
                    inner_owner.screen_name = "owner"
                    inner_owner.name = "Owner"
                    inner_owner.create()
                    inner_manager = InnerRole()
                    inner_manager.screen_name = "manager"
                    inner_manager.name = "Manager"
                    inner_manager.create()
                    inner_normal = InnerRole()
                    inner_normal.screen_name = "normal"
                    inner_normal.name = "Normal"
                    inner_normal.create()
                except Exception:
                    pass

                Permission.accept("admin", owner)
                Permission.accept("create_user", owner)
                Permission.accept("update_user", owner)
                Permission.accept("delete_user", owner)
                Permission.accept("create_group", owner)
                Permission.accept("update_group", owner)
                Permission.accept("delete_group", owner)
                Permission.accept("create_user", manager)
                Permission.accept("update_user", manager)
                Permission.accept("create_group", manager)
                Permission.accept("update_group", manager)

                InnerPermission.accept("admin", inner_owner)
                InnerPermission.accept("add_user", inner_owner)
                InnerPermission.accept("remove_user", inner_owner)
                InnerPermission.accept("add_group", inner_owner)
                InnerPermission.accept("remove_group", inner_owner)
                InnerPermission.accept("add_user", inner_manager)
                InnerPermission.accept("remove_user", inner_owner)
                InnerPermission.accept("add_group", inner_manager)
                InnerPermission.accept("remove_group", inner_owner)

                template = self.view.get_template("confirm.html")
                return Response.render(template)
            except ValidationError as err:
                return Response.render(template, {"error": str(err)})
        return Response.render(template)


class HomeController(Controller):
    def handle(self, request):
        template = self.view.get_template('home.html')
        try:
            with open(self.app.project_dir / "welcome_message.md", "r") as f:
                welcome_message = f.read()
        except Exception:
            welcome_message = """
# Mitamaへようこそ

このページに社内システムの使い方を記述しておきましょう！
            """
        return Response.render(template, {
            "welcome_message": welcome_message
        })
    def settings(self, request):
        template = self.view.get_template("settings.html")
        try:
            with open(self.app.project_dir / "welcome_message.md", "r") as f:
                welcome_message = f.read()
        except Exception:
            welcome_message = """
# Mitamaへようこそ

このページに社内システムの使い方を記述しておきましょう！
            """
        if request.method == "POST":
            try:
                form = SettingsForm(request.post())
                welcome_message_ = form["welcome_message"]
                if form["role_screen_name"] is not None:
                    role = Role()
                    role.screen_name = form['role_screen_name']
                    role.name = form['role_name'] if form['role_name'] is not None else form['role_screen_name']
                    role.create()
                for permission_screen_name, permission_roles in form['permission'].items():
                    permission = Permission.retrieve(screen_name = permission_screen_name)
                    permission.roles = [Role.retrieve(screen_name=role) for role in permission_roles]
                    permission.update()
                if form["inner_role_screen_name"] is not None:
                    inner_role = InnerRole()
                    inner_role.screen_name = form['inner_role_screen_name']
                    inner_role.name = form['inner_role_name'] if form['inner_role_name'] is not None else form['inner_role_screen_name']
                    inner_role.create()
                for inner_permission_screen_name, inner_permission_roles in form['inner_permission'].items():
                    inner_permission = InnerPermission.retrieve(screen_name=inner_permission_screen_name)
                    inner_permission.roles = [InnerRole.retrieve(screen_name=role) for role in inner_permission_roles]
                    inner_permission.update()
                with open(self.app.project_dir / "welcome_message.md", "w") as f:
                    f.write(welcome_message_)
                error = "変更を保存しました"
            except ValidationError as err:
                error = str(err)
            welcome_message = welcome_message_
            return Response.render(template, {
                "welcome_message": welcome_message,
                'roles': Role.list(),
                'permissions': Permission.list(),
                'inner_roles': InnerRole.list(),
                'inner_permissions': InnerPermission.list(),
                "error": error
            })
        return Response.render(template, {
            "welcome_message": welcome_message,
            'roles': Role.list(),
            'permissions': Permission.list(),
            'inner_roles': InnerRole.list(),
            'inner_permissions': InnerPermission.list(),
        })


class UsersController(Controller):
    def create(self, req):
        template = self.view.get_template("user/create.html")
        invites = User.query.filter(User.password==None).all()
        if req.method == "POST":
            try:
                form = InviteForm(req.post())
                user = UserInvite()
                user.email = form["email"]
                user.name = form["name"]
                user.screen_name = form["screen_name"]
                user._icon = resize_icon(form["icon"])
                user.roles = ":".join(form["roles"])
                user.create()
                user.mail(
                    "Mitamaに招待されています",
                    "下記リンクから、Mitamaに参加しましょう\n{}".format(self.app.convert_fullurl(req, "/signup?token=" + user.token))
                )
                return Response.render(
                    template,
                    {
                        "invites": invites,
                        "roles": Role.list(),
                        "icon": load_noimage_user()
                    }
                )
            except Exception as err:
                error = str(err)
                return Response.render(
                    template,
                    {
                        "invites": invites,
                        "roles": Role.list(),
                        "name": form["name"],
                        "screen_name": form["screen_name"],
                        "icon": resize_icon(form["icon"]),
                        "error": error,
                    },
                )
        return Response.render(
            template,
            {
                "invites": invites,
                "roles": Role.list(),
                "icon": load_noimage_user()
            }
        )

    def cancel(self, req):
        invite = Invite.retrieve(req.params["id"])
        invite.delete()
        return Response.redirect(self.app.convert_url("/users/invite"))

    def retrieve(self, req):
        template = self.view.get_template("user/retrieve.html")
        user = User.retrieve(screen_name=req.params["id"])
        return Response.render(
            template,
            {
                "user": user,
            },
        )

    def update_profile(self, req):
        template = self.view.get_template("user/update/profile.html")
        user = User.retrieve(screen_name=req.params["id"])
        roles = Role.list()
        error = ""
        if req.method == "POST":
            form = UserUpdateForm(req.post())
            try:
                user.screen_name = form["screen_name"]
                user.name = form["name"]
                user.icon = resize_icon(form["icon"]) if form["icon"] is not None else user.icon
                roles_ = []
                for role in form["roles"]:
                    roles_.append(Role.retrieve(screen_name=role))
                user.roles = roles_
                user.update()
                return Response.redirect(self.app.convert_url("/users/" + user.screen_name + "/settings"))
            except Exception as err:
                error = str(err)
        return Response.render(
            template,
            {
                "user": user,
                "screen_name": user.screen_name,
                "name": user.name,
                "icon": user.icon,
                "roles": roles,
                "error": error
            },
        )

    def update_password(self, req):
        template = self.view.get_template("user/update/password.html")
        user = User.retrieve(screen_name=req.params["id"])
        if user._id != req.user._id:
            return Response.redirect(self.app.convert_url("/"))
        error = ""
        if req.method == "POST":
            form = UserPasswordUpdateForm(req.post())
            try:
                if form["password"] != form["password_"]:
                    raise Exception("パスワードが一致しません")
                user.set_password(form["password"])
                user.update()
                error = "パスワードを変更しました"
            except Exception as err:
                error = str(err)
        return Response.render(
            template,
            {
                "user": user,
                "error": error
            },
        )

    def update_notification(self, req):
        template = self.view.get_template("user/update/notification.html")
        user = User.retrieve(screen_name=req.params["id"])
        if user._id != req.user._id:
            return Response.redirect(self.app.convert_url("/"))
        error = ""
        if req.method == "POST":
            form = SubscriptionForm(req.post())
            try:
                if form["action"] == "subscribe":
                    subscription = PushSubscription()
                    subscription.subscription = form["subscription"]
                    subscription.create()
                    user.subscriptions.append(subscription)
                    user.update()
                    error="通知の購読を設定しました"
                else:
                    subscription = PushSubscription.retrieve(subscription = form["subscription"])
                    subscription.delete()
                    error="通知の購読を解除しました"
            except Exception as err:
                error = str(err)
        return Response.render(
            template,
            {
                "user": user,
                "vapid_public_key": self.app.project.vapid["public_key"],
                "error": error
            },
        )

    def delete(self, req):
        template = self.view.get_template("user/delete.html")
        return Response.render(template)

    def list(self, req):
        template = self.view.get_template("user/list.html")
        users = User.list()
        return Response.render(
            template,
            {
                "users": users,
            },
        )


class GroupsController(Controller):
    def create(self, req):
        template = self.view.get_template("group/create.html")
        groups = Group.list()
        error = ""
        if req.method == "POST":
            form = GroupCreateForm(req.post())
            try:
                group = Group()
                group.name = form["name"]
                group.screen_name = form["screen_name"]
                group.icon = resize_icon(form["icon"])
                group.create()
                if form["parent"] != None:
                    Group.retrieve(form["parent"]).append(group)
                group.append(req.user)
                return Response.redirect(self.app.convert_url("/groups"))
            except ValidationError as err:
                error = str(err)
        return Response.render(
            template, {
                "groups": groups,
                "icon": load_noimage_group(),
                "error": error
            }
        )

    def retrieve(self, req):
        template = self.view.get_template("group/retrieve.html")
        group = Group.retrieve(screen_name=req.params["id"])
        return Response.render(
            template,
            {
                "group": group,
            },
        )

    def update(self, req):
        template = self.view.get_template("group/update.html")
        group = Group.retrieve(screen_name=req.params["id"])

        groups = list()
        for g in Group.list():
            if not (group.is_ancestor(g) or group.is_descendant(g) or g == group):
                groups.append(g)
        users = list()
        error = ""
        message = ""
        for u in User.list():
            if not group.is_in(u):
                users.append(u)
        if req.method == "POST":
            form = GroupUpdateForm(req.post())
            try:
                icon = resize_icon(form["icon"]) if form["icon"] is not None else group.icon
                group.screen_name = form["screen_name"]
                group.name = form["name"]
                group.parent = Group.retrieve(form["parent"]) if form["parent"] is not None else None
                for role in form["roles"]:
                    group.roles.append(Role.retrieve(screen_name=role))
                group.icon = icon
                group.users = [User.retrieve(user) for user in form['users']]
                if form['new_user'] is not None:
                    group.users.append(User.retrieve(form['new_user']))
                for user, roles in form['inner_roles'].items():
                    rel = UserGroup.retrieve(group = group, user = User.retrieve(user))
                    rel.roles = [InnerRole.retrieve(screen_name = role) for role in roles]
                group.update()
                message = "変更を保存しました"
            except ValidationError as err:
                error = str(err)
        roles = Role.list()
        inner_roles = InnerRole.list()
        return Response.render(
            template,
            {
                "message": message,
                "error": error,
                "group": group,
                "all_groups": groups,
                "all_users": users,
                "screen_name": group.screen_name,
                "name": group.name,
                "icon": group.icon,
                "roles": roles,
                "inner_roles": inner_roles
            },
        )

    def append(self, req):
        form = req.post()
        try:
            group = Group.retrieve(screen_name=req.params["id"])
            nodes = list()
            if "user" in form:
                for uid in form.getlist("user"):
                    try:
                        nodes.append(User.retrieve(uid))
                    except Exception as err:
                        pass
            if "group" in form:
                for gid in form.getlist("group"):
                    try:
                        nodes.append(Group.retrieve(gid))
                    except Exception as err:
                        pass
            group.append_all(nodes)
        except Exception:
            pass
        finally:
            return Response.redirect(
                self.app.convert_url("/groups/" + group.screen_name + "/settings")
            )

    def remove_user(self, req):
        try:
            group = Group.retrieve(screen_name=req.params["id"])
            child = User.retrieve(req.params["uid"])
            group.remove(child)
        except Exception as err:
            print(err)
            pass
        finally:
            return Response.redirect(
                self.app.convert_url("/groups/" + group.screen_name + "/settings")
            )

    def remove_group(self, req):
        try:
            group = Group.retrieve(screen_name=req.params["id"])
            child = Group.retrieve(req.params["gid"])
            group.remove(child)
        except Exception as err:
            print(err)
            pass
        finally:
            return Response.redirect(
                self.app.convert_url("/groups/" + group.screen_name + "/settings")
            )

    def accept(self, req):
        group = Group.retrieve(screen_name=req.params["id"])
        user = User.retrieve(req.params["uid"])
        inner_role = InnerRole.retrieve(screen_name = "admin")
        inner_role.append(group, user)
        return Response.redirect(
            self.app.convert_url("/groups/" + group.screen_name + "/settings")
        )

    def forbit(self, req):
        group = Group.retrieve(screen_name=req.params["id"])
        user = User.retrieve(req.params["uid"])
        inner_role = InnerRole.retrieve(screen_name = "admin")
        inner_role.remove(group, user)
        return Response.redirect(
            self.app.convert_url("/groups/" + group.screen_name + "/settings")
        )

    def delete(self, req):
        template = self.view.get_template("group/delete.html")
        return Response.render(template)

    def list(self, req):
        template = self.view.get_template("group/list.html")
        groups = Group.tree()
        return Response.render(
            template,
            {
                "groups": groups,
            },
        )


class AppsController(Controller):
    def list(self, req):
        template = self.view.get_template("apps/list.html")
        apps = AppRegistry()
        return Response.render(
            template,
            {
                "apps": apps,
            },
        )


class ACSController(Controller):
    def redirect(request):
        pass

    def post(request):
        pass


class SLOController(Controller):
    def redirect(request):
        pass

    def post(request):
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mitama',
 'mitama.app',
 'mitama.app.forms',
 'mitama.app.http',
 'mitama.db',
 'mitama.db.driver',
 'mitama.models',
 'mitama.portal',
 'mitama.project',
 'mitama.skeleton.app_templates',
 'mitama.skeleton.project',
 'mitama.utils']

package_data = \
{'': ['*'],
 'mitama.app': ['static/*', 'templates/*', 'templates/base/*'],
 'mitama.portal': ['static/*',
                   'templates/*',
                   'templates/apps/*',
                   'templates/group/*',
                   'templates/user/*',
                   'templates/user/update/*'],
 'mitama.skeleton.app_templates': ['static/*', 'templates/*']}

install_requires = \
['Jinja2>=2.11.3,<3.0.0',
 'Markdown>=3.3.3,<4.0.0',
 'Pillow>=8.1.0,<9.0.0',
 'PyJWT>=2.0.1,<3.0.0',
 'SQLAlchemy>=1.3.23,<2.0.0',
 'bcrypt>=3.2.0,<4.0.0',
 'gevent-websocket>=0.10.1,<0.11.0',
 'gevent>=21.1.2,<22.0.0',
 'jinja-markdown>=1.200630,<2.0',
 'pycryptodome>=3.10.1,<4.0.0',
 'pysaml2>=6.5.1,<7.0.0',
 'python-magic>=0.4.22,<0.5.0',
 'pywebpush>=1.11.0,<2.0.0',
 'tzlocal>=2.1,<3.0',
 'uWSGI>=2.0.19,<3.0.0',
 'watchdog>=2.0.1,<3.0.0']

entry_points = \
{'console_scripts': ['mitama = mitama:command_exec']}

setup_kwargs = {
    'name': 'mitama',
    'version': '4.5.11',
    'description': '',
    'long_description': None,
    'author': 'boke0',
    'author_email': 'speken00.tt@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from . import errors
import re

class Field:
    ValidationError = errors.ValidationError
    EmptyError = errors.EmptyError
    FormatError = errors.FormatError
    def __init__(
        self,
        label="",
        initial=None,
        regex=None,
        validator=None,
        required=False,
        name=None,
        form_type="text",
        placeholder=None,
        listed=False
    ):
        self.required = required
        self.label = label
        self.initial = initial
        self.regex = regex
        self.validator = validator
        self.form_type = form_type
        self.placeholder = placeholder
        self.name = name
        self.listed = listed

    def instance(self):
        return FieldInstance(
            label=self.label,
            initial=self.initial,
            regex=self.regex,
            validator=self.validator,
            required=self.required,
            name=self.name,
            form_type=self.form_type,
            placeholder=self.placeholder,
            listed=self.listed,
            validation_error=self.ValidationError,
            empty_error=self.EmptyError,
            format_error=self.FormatError,
        )

class DictField(Field):
    def __init__(
        self,
        **kwargs
    ):
        super().__init__(
            **kwargs
        )

    def instance(self):
        return DictFieldInstance(
            label=self.label,
            initial=self.initial,
            regex=self.regex,
            validator=self.validator,
            required=self.required,
            name=self.name,
            form_type=self.form_type,
            placeholder=self.placeholder,
            listed=self.listed,
            validation_error=self.ValidationError,
            empty_error=self.EmptyError,
            format_error=self.FormatError,
        )

class FileField(Field):
    def __init__(
        self,
        accept=None,
        max_file_size=None,
        **kwargs
    ):
        kwargs["form_type"] = "file"
        super().__init__(
            **kwargs
        )
        self.accept = None
        self.max_file_size = None

    def instance(self):
        return FileFieldInstance(
            required=self.required,
            label=self.label,
            initial=self.initial,
            validator=self.validator,
            name=self.name,
            placeholder=self.placeholder,
            accept=self.accept,
            max_file_size=self.max_file_size,
            validation_error=self.ValidationError,
            empty_error=self.EmptyError,
            format_error=self.FormatError,
        )

class FieldInstance:
    def __init__(
        self,
        label="",
        initial=None,
        regex=None,
        validator=None,
        required=False,
        name=None,
        form_type="text",
        placeholder=None,
        listed=False,
        validation_error=errors.ValidationError,
        empty_error=errors.EmptyError,
        format_error=errors.FormatError,
    ):
        self.required = required
        self.label = label
        self.initial = initial
        self.regex = regex
        self.validator = validator
        self.data = initial
        self.form_type = form_type
        self.placeholder = placeholder
        self.name = name
        self.listed=listed
        self.validation_error = validation_error
        self.empty_error = empty_error
        self.format_error = format_error

    def reset(self):
        self.data = self.initial

    def validate(self):
        if self.required and self.data is None:
            raise self.empty_error(self.label)
        if self.data is not None and self.regex and re.fullmatch(self.regex, self.data) is None:
            raise self.format_error(self.label, self.data)
        if self.data is not None and self.validator: self.validator(self.data)
        return True

class DictFieldInstance:
    def __init__(
        self,
        label="",
        initial=None,
        regex=None,
        validator=None,
        required=False,
        name=None,
        form_type="text",
        placeholder=None,
        listed=False,
        validation_error=errors.ValidationError,
        empty_error=errors.EmptyError,
        format_error=errors.FormatError,
    ):
        self.required = required
        self.label = label
        self.initial = initial
        self.regex = regex
        self.validator = validator
        self.data = initial
        self.form_type = form_type
        self.placeholder = placeholder
        self.name = name
        self.listed=listed
        self.validation_error = validation_error
        self.empty_error = empty_error
        self.format_error = format_error

    def reset(self):
        self.data = self.initial

    def validate(self):
        if self.required and self.data is None:
            raise self.empty_error(self.label)
        if self.data is not None and self.regex and re.fullmatch(self.regex, self.data) is None:
            raise self.format_error(self.label, self.data)
        if self.data is not None and self.validator: self.validator(self.data)
        return True

class FileFieldInstance(FieldInstance):
    def __init__(
        self,
        label="",
        initial=None,
        regex=None,
        validator=None,
        required=False,
        name=None,
        accept=None,
        placeholder=None,
        max_file_size=None,
        validation_error=errors.ValidationError,
        empty_error=errors.EmptyError,
        format_error=errors.FormatError,
    ):
        super().__init__(
            required=required,
            label=label,
            initial=initial,
            validator=validator,
            name=name,
            placeholder=placeholder,
            form_type="file",
            validation_error=validation_error,
            empty_error=empty_error,
            format_error=format_error,
        )
        self.accept = accept
        self.max_file_size = max_file_size
        self._data = None
        self._bin = None

    def validate(self):
        if self.required and self.data is None:
            raise self.empty_error(self.label)
        if self.data is not None and self.regex and re.fullmatch(self.regex, self.data) is None:
            raise self.format_error(self.label, self.data)
        if self.data is not None and self.validator: self.validator(self.data)
        return True

    @property
    def data(self):
        if self._data is not None:
            if self._bin is None:
                self._bin = self._data.file.read()
            if self._bin is not None or len(self._bin) > 0:
                return self._bin
            else:
                return None
        elif self.initial is not None:
            return self.initial
        else:
            return None

    @data.setter
    def data(self, value):
        self._data = value

